/* tsx.c                      tab  November 2013 */

/* Process TripleSpec raw data files. */

#include "fitsio.h"
#include "tsx.h"
#include "mua.h"
#include "cfua.h"
#include "galactic.h"

/* original:
#define BLUE2RED 4.972
*/
#define BLUE2RED 3.420

/* DBSP Specs:

  2.57  arcsec/mm at the slit
  0.293 arcsec/pixel for red camera (new) 
  0.389 arcsec/pixel for blue camera      [ 0.151 mm/pix ]

New LBNL Red: 4096x2048; 15micron pixels; linear to 40,000 DN; saturation 45,000 DN
        Blue: 2048x4096; 15micron pixels; linear to 62,000 DN; saturation 65,000 DN (575-3410 unvig.);

slit length = 128 arcsec.. widths: 0.5 1.0 1.5 2.0 4.0 6.0 8.0 10.0 arcsec

                          (dispersion Ang/mm)
     lines/mm / blaze(A)    red     blue
gratings:  600/4000                  72

blue 600/4000:  72 A/mm x 0.151 mm/pix = 10.87 Ang/pixel .. (not correct!) WRONG...


*/


/* TripleSpec LiSting of exposures. */
#define MAXTLS 900
struct TLSstructure {
  int num;
  char xtype[9];
  char root[30];
  double ra,dec,air,jd,exp;
  double EffTemp;
};
typedef struct TLSstructure TLStype;



/* Tycho Star Catalog */
#define MAXTYC 2701000
struct TYCstructure {
  double mean_ra,mean_dec;
  float bt_mag,vt_mag;
  float bt_mag_error,vt_mag_error;
};
typedef struct TYCstructure TYCtype;




/* Triple Spec Polynomial solutions to edges of orders. */
struct TSPstructure {
  double xoffa;
  double coefa[3];
  double xoffb;
  double coefb[3];
};
typedef struct TSPstructure TSPtype;

/* Triple Spec Polynomial solutions to offsets between orders. */
struct TOFFstructure {
  int polyorder;
  double ps_xoff;
  double ps_coef[5];
};
typedef struct TOFFstructure TOFFtype;

/* Inverse: Triple Spec Polynomial solutions to offsets between orders. */
struct TOFFIstructure {
  int polyorder;
  double ps_xoff;
  double ps_coef[5];
};
typedef struct TOFFIstructure TOFFItype;



/* Triple-spec Order Edges */
struct TOEstructure {
  int num;            /* TOE[0].num is number of entries. */
  int column;         /* Column number for this entry. */
  int nedge;          /* Number of edges at this column. */
  int edge1[9];       /* Lower edge (row number) of an order. */
  int edge2[9];       /* Upper edge (row number) of an order. */
};
typedef struct TOEstructure TOEtype;

#define VERB 0
#define READ_DARKIMG 0

#define EDGEFIND 0

/* ORIGINAL:
#define EPERDN   4.49   /x default is 3.8 in cookbook x/
#define ROVDN    2.51   /x implies 11.3 electrons: cookbook says 10e(CDS), 5e(8 samples), 3.5e(16 samples) x/
#define RAWOFF  10.00
*/

/* REFINED based on extracting sky lines...  */
#define EPERDN   3.50
#define ROVDN    4.00
#define RAWOFF -10.00




/* ----------------------------------------------------------------------
  Return the value of the black body function in ergs/sec/cm^2/Angstrom
  (i.e. energy per unit wavelength):
      flux(a,T,x) = a * ( (h*c*c)/x^5 ) / ( exp((h*c)/(k*T*x)) - 1 )
  Input:  a    : Arbitrary scale factor.
          T    : Temperature in Kelvin.
          x    : Wavelength in Angstroms.
*/
/*@@*/
double tsx_BlackBodyAT( double a, double T, double x )
{
/**/
  const double c = 2.997925e+18 ;  /*  Ang/sec      */
  const double k = 1.380662e-16 ;  /*  ergs/Kelvin  */
  const double h = 6.626176e-27 ;  /*  ergs*sec     */
  double p1 = h*c*c;
  double p2 = h*c/k;
/**/
  double r;
/**/
  r = ( a * p1  / (x*x*x*x*x) ) / ( exp(p2/(T*x)) - 1. ) ;
  return(r);
}

/* ----------------------------------------------------------------------
  Given a temperature and a weighted spectrum, return the best possible fit
  for the value "a" using:
     flux(a,T,x) = a * ( (h*c*c)/x^5 ) / ( exp((h*c)/(k*T*x)) - 1 )
  Use:
     a = SUM { Wi * Yi * flux(1,T,Xi) } / SUM { Wi * flux(1,T,Xi)^2 }
  Input: nn    : number of points in spectrum.
         xx[]  : wavelengths in Angstroms.
         yy[]  : flux in ergs/sec/cm^2/Angstroms.
         ww[]  : weight values.
         T     : Temperature in Kelvin.
*/
/*@@*/
double tsx_SolveBlackBodyA( int nn, float xx[], float yy[], float ww[], double T )
{
/**/
double a,sum1,sum2,x,y,w,flux;
int ii;
/**/
a   =1.;
sum1=0.;
sum2=0.;
for(ii=0; ii<nn; ++ii) {
  x = xx[ii];
  y = yy[ii];
  w = ww[ii];
  flux = tsx_BlackBodyAT(a,T,x);
  sum1 = sum1 + ( w * y * flux );
  sum2 = sum2 + ( w * flux * flux );
}
if (sum2 > 0.) { return(sum1/sum2); } else { return(0.); }
}

/* ----------------------------------------------------------------------
  Given a spectrum, and black-body fit parameters, return the least-squares
  value.
  Input: nn    : number of points in spectrum.
         xx[]  : wavelengths in Angstroms.
         yy[]  : flux in ergs/sec/cm^2/Angstroms.
         ww[]  : weight values.
  Input: T     : Temperature in Kelvin.
         a     : Temperature in Kelvin.
*/
/*@@*/
double tsx_LeastSquaresBlackBody( int nn, float xx[], float yy[], float ww[], 
                               double T, double a )
{
/**/
double sum,x,y,w,r;
int ii;
/**/
sum = 0.;
for (ii=0; ii<nn; ++ii) {
  x = xx[ii];
  y = yy[ii];
  w = ww[ii];
  r = (tsx_BlackBodyAT(a,T,x) - y);  
  sum = sum + (w * r * r);
}
return(sum);
}

/* ----------------------------------------------------------------------
  Given a spectrum, estimate the black-body temperature "T" and the scale
  parameter "a", assuming the flux is in ergs/sec/cm^2/Angstrom
  (i.e. energy per unit wavelength):
       flux(a,T,x) = a * ( (h*c*c)/x^5 ) / ( exp((h*c)/(k*T*x)) - 1 )
  Input: nn    : number of points in spectrum.
         xx[]  : wavelengths in Angstroms.
         yy[]  : flux in ergs/sec/cm^2/Angstroms.
         ww[]  : weight values.
 Output: T     : Temperature in Kelvin.
         a     : Temperature in Kelvin.
*/
/*@@*/
void tsx_FitBlackBody( int nn, float xx[], float yy[], float ww[], 
                       double *T, double *a )
{
/**/
double tt,aa,lsf,lo_lsf,lo_aa,lo_tt;
/**/
/* Set up. */
lo_tt = 6000.;
lo_aa = tsx_SolveBlackBodyA( nn, xx, yy, ww, lo_tt );
lo_lsf= tsx_LeastSquaresBlackBody( nn, xx, yy, ww, lo_tt, lo_aa );
/* Find the best temperature. */
for (tt=2000.; tt<9000.; tt=tt+200.) {
  aa = tsx_SolveBlackBodyA( nn, xx, yy, ww, tt );
  lsf= tsx_LeastSquaresBlackBody( nn, xx, yy, ww, tt, aa );
  if (lsf < lo_lsf) { lo_lsf=lsf; lo_aa=aa; lo_tt=tt; }
}
for (tt=9000.; tt<21000.; tt=tt+500.) {
  aa = tsx_SolveBlackBodyA( nn, xx, yy, ww, tt );
  lsf= tsx_LeastSquaresBlackBody( nn, xx, yy, ww, tt, aa );
  if (lsf < lo_lsf) { lo_lsf=lsf; lo_aa=aa; lo_tt=tt; }
}
*T = lo_tt;
*a = lo_aa;
return;
}





/* ----------------------------------------------------------------------
    Find the nearest element in an array to the given value.
       Input: xx (given value).
       Input: narr (number of elements in array).
       Input: arr (the array to be searched).
    Array need not be sorted.
*/
int tsx_cneari(double xx, int narr, double arr[])
{
/**/
int ii,bestii;
double lo,diff;
/**/
bestii=0;
lo=ABS((xx - arr[0]));
for (ii=0; ii<narr; ++ii) {
  diff = ABS((xx - arr[ii]));
  if (diff < lo) { lo=diff; bestii=ii; }
}
return(bestii);
}


/* ----------------------------------------------------------------------
 Append a slash if needed.
*/
void tsx_append_slash( char wrd[] )
{
int ii;
ii = clc(wrd);
if (ii >= 0) { if (wrd[ii] != '/') { wrd[(ii+1)]='/';   wrd[(ii+2)]='\0'; } }
return;
}


/* ----------------------------------------------------------------------
  Write out a FITS image.
*/
void tsx_wout( char ofile[], float image[], int nc, int nr )
{
/**/
char wrd[100];
/**/
fitsfile *fptr;
/**/
int bitpix,hdutype;
int status = 0;
/**/
long nelements;
long naxis = 2;
long fpixel = 1;
long naxes[9];
/**/

/* Create a new FITS image. */
bitpix = -32;
naxes[0] = nc;
naxes[1] = nr;
strcpy(wrd,"!"); strcat(wrd,ofile);
if (VERB) printf("NOTE: Writing FITS image file '%s'.\n",ofile);
fits_create_file( &fptr, wrd, &status );                 cfua_error( status );
fits_create_img( fptr, bitpix, naxis, naxes, &status );  cfua_error( status );
fits_movabs_hdu( fptr, 1, &hdutype, &status );           cfua_error( status );

/* Write image. */
nelements = nc * nr;
fits_write_img( fptr, TFLOAT, fpixel, nelements, image, &status ); cfua_error(status);
fits_close_file( fptr, &status );                                  cfua_error(status);

return;
}


/* ----------------------------------------------------------------------
 Mash/median rows(1) or columns(2).
 Size of mash[] will be either 'nc'(mode=1) or 'nr'(mode=2).
 Mash will be between up1 and up2 (which are either row numbers(mode=1) 
 or column numbers(mode=2)).  These values are limited by bounds.
*/
void tsx_mash_median( int mode, float image[], int nc, int nr, double mash[], int up1, int up2 )
{
/**/
int pixno,ii,jj,p1,p2,narr;
/**/
float arr[9000];
/**/
if (mode == 1) {
  p1 = MIN((nr-1),MAX(0,(up1)));
  p2 = MIN((nr-1),MAX(0,(up2)));
  for (ii=0; ii<nc; ++ii) {
    narr=0;
    for (jj=p1; jj<=p2; ++jj) { 
      pixno=ii+(jj*nc); arr[narr]=image[pixno]; ++narr;
    }
    mash[ii]= cfind_median(narr,arr);
  }
}
if (mode == 2) {
  p1 = MIN((nc-1),MAX(0,(up1)));
  p2 = MIN((nc-1),MAX(0,(up2)));
  for (jj=0; jj<nr; ++jj) {
    narr=0;
    for (ii=p1; ii<=p2; ++ii) { 
      pixno=ii+(jj*nc); arr[narr]=image[pixno]; ++narr;
    }
    mash[jj]= cfind_median(narr,arr);
  }
}
return;
}


/* ----------------------------------------------------------------------
 Mash rows(1) or columns(2).
 Size of mash[] will be either 'nc'(mode=1) or 'nr'(mode=2).
 Mash will be between up1 and up2 (which are either row numbers(mode=1) 
 or column numbers(mode=2)).  These values are limited by bounds.
*/
void tsx_mash( int mode, float image[], int nc, int nr, double mash[], int up1, int up2 )
{
/**/
int pixno,ii,jj,p1,p2;
/**/
double sum,num;
/**/
if (mode == 1) {
  p1 = MIN((nr-1),MAX(0,(up1)));
  p2 = MIN((nr-1),MAX(0,(up2)));
  for (ii=0; ii<nc; ++ii) {
    sum=0.; num=0.;
    for (jj=p1; jj<=p2; ++jj) { pixno=ii+(jj*nc); sum=sum+image[pixno]; num=num+1.0; }
    mash[ii]= sum/num;
  }
}
if (mode == 2) {
  p1 = MIN((nc-1),MAX(0,(up1)));
  p2 = MIN((nc-1),MAX(0,(up2)));
  for (jj=0; jj<nr; ++jj) {
    sum=0.; num=0.;
    for (ii=p1; ii<=p2; ++ii) { pixno=ii+(jj*nc); sum=sum+image[pixno]; num=num+1.0; }
    mash[jj]= sum/num;
  }
}
return;
}


/* ----------------------------------------------------------------------
  Centroid.
*/
double tsx_centroid( int nn, double yy[], int uu1, int uu2 )
{
/**/
double cent,loyy,sum,wsum;
int ii,ii1,ii2;
/**/
ii1=uu1; ii2=uu2;
if (ii1 < 0) ii1=0;
if (ii2 > (nn-1)) ii2=nn-1;
loyy=0.;
for (ii=ii1; ii<=ii2; ++ii) { if (yy[ii] < loyy) loyy=yy[ii]; }
sum=0.; wsum=0.;
for (ii=ii1; ii<=ii2; ++ii) { 
  sum = sum + ( (double)ii * (yy[ii] - loyy) );
  wsum= wsum+ (yy[ii] - loyy);
}
if (wsum > 0.) { cent = sum / wsum; } else { printf("***error: bad centroid.\n"); exit(1); }
return(cent);
}



/* ----------------------------------------------------------------------
  Find real value of lower bound at a given column.
*/
double tsx_find_real_image_row( int icol, int tso, double polyoff[], TSPtype TSP[] )
{
/**/
double xv,yv,coef[9];
/**/
coef[0] = TSP[tso].coefa[0];
coef[1] = TSP[tso].coefa[1];
coef[2] = TSP[tso].coefa[2];
xv = (double)icol - TSP[tso].xoffa;
yv =cpolyval(3,coef,xv) + polyoff[tso];
return(yv);
}



/* ----------------------------------------------------------------------
  Find image row given a column, triple spec order, and polynomials.
  mode=1 : lower bound.
  mode=2 : upper bound.
  The lower bound is offset for some orders to align profiles.
*/
int tsx_find_image_row( int mode, int icol, int tso, double polyoff[], TSPtype TSP[] )
{
/**/
int row;
double xv,yv,coef[9];
/**/
if (mode == 1) {
  coef[0] = TSP[tso].coefa[0];
  coef[1] = TSP[tso].coefa[1];
  coef[2] = TSP[tso].coefa[2];
  xv = (double)icol - TSP[tso].xoffa;
  yv =cpolyval(3,coef,xv) + polyoff[tso];
} else {
  coef[0] = TSP[tso].coefb[0];
  coef[1] = TSP[tso].coefb[1];
  coef[2] = TSP[tso].coefb[2];
  xv = (double)icol - TSP[tso].xoffb;
  yv =cpolyval(3,coef,xv);
}
row = cnint(( yv ));
return(row);
}


/* #@# 
#define POLYOFF4 1.62 
#define POLYOFF5 3.54
#define POLYOFF6 4.61
#define POLYOFF7 5.68
  #@# */



/*----------------------------------------------------------------------
  Load TYC[]. 
  'TYC.bin' was created by some code in /home/tb/c/f/tst.c ... -tab August 2014
*/
void tsx_load_TYC( char tsxdir[], TYCtype TYC[], int *numTYC )
{
/**/
char tycfile[200];
FILE *binfu;
int nn;
/**/
sprintf(tycfile,"%scal/TYC.bin",tsxdir);
printf("Reading '%s'.\n",tycfile);
nn=0;
binfu = fopen_read(tycfile);
while ( fread((char *)&TYC[nn], 1, sizeof(TYCtype), binfu ) == sizeof(TYCtype) ) { ++nn; }
fclose(binfu);
*numTYC = nn;
printf("Read in %d Tycho entries.\n",*numTYC);
return;
}


/*----------------------------------------------------------------------
  Load calibration info.   -tab 14may2013
*/
void tsx_load_cal( char tsxdir[], TSPtype TSP[], TOFFtype TOFF[], TOFFItype TOFFI[], TYCtype TYC[], int *numTYC )
{
/**/
int mm,ii,tso,nn;
/**/
char wrd[200];
char line[200];
/**/
FILE *infu;
/**/
double coef[9],xoff,xv;
double xx[900],yy[900],ww[900];
/**/
/*
FILE *tstfu;
*/

/* Load Tycho catalog.  */
tsx_load_TYC( tsxdir, TYC, numTYC );

/* #@#
tstfu = fopen_write("tyc.out");
for (ii=0; ii<*numTYC; ++ii) {
  fprintf(tstfu," %8d %11.5f %11.5f %9.3f %9.3f \n",ii,TYC[ii].mean_ra,TYC[ii].mean_dec,TYC[ii].bt_mag,TYC[ii].vt_mag);
}
fclose(tstfu);
printf("#@# look at tyc.out\n"); cpauseit();
   #@# */

/* Load polynomials into structure. */
/* These polynomials determine the edges of the slit for each order. */
sprintf(wrd,"%scal/tsx_edge.dat",tsxdir);  
if (VERB) printf("Read '%s'.\n",wrd);
infu = fopen_read(wrd);
for (tso=3; tso<=7; ++tso) {
  fgetline(line,infu);
  TSP[tso].xoffa=GLV(line,2); TSP[tso].coefa[0]=GLV(line,3); TSP[tso].coefa[1]=GLV(line,4); TSP[tso].coefa[2]=GLV(line,5);
  fgetline(line,infu);
  TSP[tso].xoffb=GLV(line,2); TSP[tso].coefb[0]=GLV(line,3); TSP[tso].coefb[1]=GLV(line,4); TSP[tso].coefb[2]=GLV(line,5);
}
fclose(infu);

/* Load spatial offset polynomials into structure. */
/* Note that the coefs for tso=3 are an arcsecond solution (relative arcsec vs. row) and
 * the rest are relative row offsets (row in tso=3 frame vs. row in tso=N). */
sprintf(wrd,"%scal/toff.poly",tsxdir);  
if (VERB) printf("Read '%s'.\n",wrd);
infu = fopen_read(wrd);
for (tso=3; tso<8; ++tso) {
  fgetline(line,infu);
  TOFF[tso].polyorder = 2;
  TOFF[tso].ps_xoff   = GLV(line,1);
  TOFF[tso].ps_coef[0]= GLV(line,2);
  TOFF[tso].ps_coef[1]= GLV(line,3);
  TOFF[tso].ps_coef[2]= GLV(line,4);
}
fclose(infu);

/* Inverse TOFF polynomials. */
for (tso=3; tso<8; ++tso) {
  nn=0;
  for (mm=0; mm<300; ++mm) {
    xx[nn] = (double)mm;
    xv     = (double)mm - TOFF[tso].ps_xoff;
    yy[nn] = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
    ww[nn] = 1.0;
    ++nn;
  }
  TOFFI[tso].polyorder = 4;
  if (GJ_polyfit(nn,yy,xx,ww,TOFFI[tso].polyorder,0,&xoff,coef) != 1) { printf("***error:%d:inverse TOFF fit failed.\n",tso); exit(1); }
  TOFFI[tso].ps_xoff = xoff;
  for (ii=0; ii<9; ++ii) { TOFFI[tso].ps_coef[ii] = coef[ii]; }

/* #@#
  for (mm=0; mm<300; mm=mm+10) {
    xv    = (double)mm - TOFF[tso].ps_xoff;
    rr1   = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
    printf("#@# for: mm=%6d  xv=%9.3f  rr1=%9.3f\n",mm,xv,rr1);
    xv    = rr1 - TOFFI[tso].ps_xoff;
    rr2   = cpolyval( TOFFI[tso].polyorder+1, TOFFI[tso].ps_coef, xv );
    printf("#@# inv: xv=%9.3f  rr2=%9.3f ..  diff=%20.9f\n",xv,rr2,(rr2 - (double)mm));
  }
  cpauseit();
*/

}



return;
}


/*----------------------------------------------------------------------
  CR correction.  -tab 31may2013  .. not working yet... -tab 22jan2014
*/
void tsx_CR_correction( int nc, int nr, float image[], float bckimg[], float bcksub[], float bckmdn[] )
{
/**/
int ii,jj,pixno,narr;
int iii,jjj,ii1,ii2,jj1,jj2;
float rr,arr[4000];
/**/
/* First do a wide median smoothing. */
for (ii=0; ii<nc; ii=ii+20) {  if (ii%1000 == 0) printf("col=%d\n",ii); 
for (jj=0; jj<nr; jj=jj+20) {
  ii1=(ii+10)-30; ii2=(ii+10)+30; jj1=(jj+10)-30; jj2=(jj+10)+30;
  if (ii1 <   0) ii1=0;
  if (ii2 >= nc) ii2=nc-1;
  if (jj1 <   0) jj1=0;
  if (jj2 >= nr) jj2=nr-1;
  narr=0;
  for (iii=ii1; iii<=ii2; ++iii) {
  for (jjj=jj1; jjj<=jj2; ++jjj) {
    pixno = iii + (jjj * nc);  
    arr[narr] = bcksub[pixno];
    ++narr;
  }}
  rr = cfind_median(narr,arr);
  for (iii=ii; iii<(ii+20); ++iii) {
  for (jjj=jj; jjj<(jj+20); ++jjj) {
    if ((iii < nc)&&(jjj < nr)) {
      pixno = iii + (jjj * nc);  
      bckmdn[pixno] = rr;
    }
  }}
}}
tsx_wout( "bm.fits", bckmdn, nc, nr );
printf("#@# wrote bm.fits.\n"); cpauseit();
return;
}


/*----------------------------------------------------------------------
Command line syntax.
*/
void tsx_syntax()
{

printf("Syntax: tsx (raw TripleSpec FITS files)  [ TripleSpec B file ] [ sp=pix,pix ] [ asp=arcsec,arcsec ]\n");
printf("\n");
printf("  The optional 'B file' is another raw FITS file where the object has\n");
printf("    been moved to another position on the slit.\n");
printf("  If 1st argument ends with '.ls' then program assumes this is a file with\n");
printf("    a list of FITS files.\n");
printf("  Create a 'tsx.tbl' exposure listing using 'tsx file.ls list=tsx.tbl', where\n");
printf("    'file.ls' is a listing of all raw TSPEC FITS files.\n");
printf("\n");
printf("Options:\n");
printf("  sp=    : Specify region to spectrally extract (pixels on reddest order#3).\n");
printf("  asp=   : Specify region to spectrally extract in arcseconds (any order).\n");
printf("  list=  : Specify a file to which to write a 'table' listing of exposures.\n");
printf("  -autox : Automatically select extraction range and extract spectrum.\n");
printf("  -autob : Automatically select a B file (if a 'tsx.tbl' file exists).\n");
printf("  -noplot: Do not create PostScript plots (use if tsxplot not created).\n");
printf("  flat=  : Normalized flat field.\n");
printf("  bk=    : Specify region for background (pixels on reddest order #3).\n");
printf("  abk=   : Specify region for background in arcseconds (any order).\n");
printf("           (bk= and abk= options may be repeated)\n");
printf("  -rbi   : Write out re-binned image with variance and wavelengths.\n");
printf("  ac=    : Root (e.g. 'tspec0045') of a Vega-type star for atmospheric correction.\n");
printf("[ Version: 13-October-2014 ]\n");
printf("\n");

return;
}


/* ----------------------------------------------------------------------
 Median filter.   -tab 03apr2014
*/
void tsx_median_filter( int bin, int nn, double arr[], double mfarr[] )
{
/**/ 
int hbin,ii,ii1,ii2,jj,tnn;
/**/ 
const int mtarr=29;
double tarr[mtarr];
double median;
/**/ 
if (bin > mtarr-2) { printf("***error: bin too large.\n"); exit(1); }
hbin=bin/2;
for (ii=0; ii<nn; ++ii) {
  ii1 = ii-hbin;
  ii2 = ii+hbin;
  if (ii1 < 0   ) ii1=0;
  if (ii2 > nn-1) ii2=nn-1;
  tnn=0;
  for (jj=ii1; jj<=ii2; ++jj) { tarr[jj-ii1] = arr[jj]; ++tnn; }
  median = cfind_median8(tnn,tarr); 
  mfarr[ii] = median;
}
return;
}



/* ----------------------------------------------------------------------
 TSX edge finder for Triple Spec flat fields.   -tab 26sep2013

      struct TOEstructure { /x Triple-spec Order Edges x/
        int num;            /x TOE[0].num is number of entries. x/
        int column;         /x Column number for this entry. x/
        int edge1[9];       /x Lower edge (row number) of an order. x/
        int edge2[9];       /x Upper edge (row number) of an order. x/
      };
      typedef struct TOEstructure TOEtype;

*/
void tsx_edge_finder( float image[], int nc, int nr, double xoffa[9], double xoffb[9], 
                      double coefa[9][9], double coefb[9][9] )
{
/**/ 
int tt,icc,ntoe,narr,ndif,ii,ii1,ii2,jj;
int nedge1,nedge2,edge1[9],edge2[9],idif[5000];
int order,tso,nn,kk,kk1,kk2,ccc,icc1[9],icc2[9];
/**/ 
double highdif[5000],meandif[5000],rmsdif[5000],dif[5000],sum1,sum2;
double thres,high,mean,sum,num,mash[5000];
double xv,ff,xx[900],yy[900],ww[900];
/**/ 
float rms,arr[2000];
/**/ 
char wrd[300];
/**/ 
FILE *outfu;
/**/ 
TOEtype TOE[1000];
/**/

/* Mash at various column intervals. */
TOE[0].num = 0;
ntoe = TOE[0].num;

/* Do in two passes. */
icc1[0]=20; icc2[0]=980;
icc1[1]=1150; icc2[1]=2010;
for (ccc=0; ccc<2; ++ccc) {

for (icc=icc1[ccc]; icc<icc2[ccc]; icc=icc+40) { 

  TOE[ntoe].column = icc;
  tsx_mash_median( 2, image, nc, nr, mash, icc-19, icc+19 );

/* Normalize. */
  sum=0.; num=0.;
  for (jj=0; jj<nr; ++jj) { sum=sum+mash[jj]; num=num+1.; }
  for (jj=0; jj<nr; ++jj) { mash[jj] = mash[jj] / (sum/num); }

/* #@#
  outfu = fopen_write("mash1.tbl");
  fprintf(outfu,"| row  | mash       |\n");
  for (jj=0; jj<nr; ++jj) { fprintf(outfu," %6d %12.5e \n",jj,mash[jj]); }
  fclose(outfu);
*/

/* Generate dif[] array. */
  ndif=0;
  for (ii=0; ii<nr-5; ++ii) {
    ii1=ii-5; ii2=ii-2;
    sum1=0; for (jj=ii1; jj<(ii2+1); ++jj) { sum1=sum1+mash[jj]; }
    ii1=ii+2; ii2=ii+5;
    sum2=0; for (jj=ii1; jj<(ii2+1); ++jj) { sum2=sum2+mash[jj]; }
    dif[ndif] = (sum1/4.) - (sum2/4.);
    idif[ndif]= ii;
    ++ndif;
  }

/* Local RMS in dif array. */
  high=0.15;
  for (ii=0; ii<ndif; ++ii) {
    if (ABS(dif[ii]) > high) high=ABS(dif[ii]);
    kk1=ii-20; kk2=ii+20; 
    if (kk1 <    0  ) kk1=0;
    if (kk2 > ndif-1) kk2=ndif-1;
    narr=0;
    for (kk=kk1; kk<kk2; ++kk) { 
      if (ABS(dif[kk]) < 0.1) {
        arr[narr]=dif[kk]; 
        ++narr;
      }
    }
    if (narr < 3) { printf("***error: not enough points for RMS of dif.\n"); exit(1); }
    mean = cfind_mean_rms( narr, arr, &rms );
    rmsdif[ii] =rms;
    meandif[ii]=mean;
    highdif[ii]=high;
  }

/* Write dif[] array. */
/* #@#
  sprintf(wrd,"dif-%4.4d.tbl",icc);  printf("Writing '%s'.\n",wrd); 
  outfu = fopen_write(wrd);
  fprintf(outfu,"| row  | absdif     | rmsdif     | meandif    | highdif    |\n");
  for (ii=0; ii<ndif; ++ii) { 
    fprintf(outfu," %6d %12.5e %12.5e %12.5e %12.5e \n",idif[ii],dif[ii],rmsdif[ii],meandif[ii],highdif[ii]/7.); 
  }
  fclose(outfu);
*/


/* Find dips, then peaks. */
  ii=200; nedge1=0; nedge2=0;
  while (ii < ndif-1) {
    thres = highdif[ii] / 7.;
    if (dif[ii] < (-0.95*thres)) {
      while (dif[ii+1] < dif[ii]) { ++ii; }
      edge1[nedge1] = idif[ii];
      ++nedge1;
      while ((ii < ndif-1)&&(dif[ii] < (-1.*thres))) { ++ii; }
    }
    if (ii < ndif-1) {
      if (dif[ii] > (0.95*thres)) {
        while (dif[ii+1] > dif[ii]) { ++ii; }
        edge2[nedge2] = idif[ii];
        ++nedge2;
        while ((ii < ndif-1)&&(dif[ii] > (thres))) { ++ii; }
      }
    }
    ++ii;
  }

/*
  printf("\n");
  for (ii=0; ii<nedge1; ++ii) { printf("#@# ii=%3d  edge1=%d\n",ii,edge1[ii]); }
  for (ii=0; ii<nedge2; ++ii) { printf("#@# ii=%3d  edge2=%d\n",ii,edge2[ii]); }
*/

  if (nedge1 != nedge2) { printf("***error: number of edges do not match, icc=%d\n",icc); exit(1); }
  TOE[ntoe].nedge = nedge1;

/* Load. */
  for (ii=0; ii<TOE[ntoe].nedge; ++ii) {
    TOE[ntoe].edge1[ii] = edge1[ii] + 3;
    TOE[ntoe].edge2[ii] = edge2[ii] - 3;
/*
    printf("#@# .. icc=%5d  edge1=%5d   edge2=%5d\n",icc,TOE[ntoe].edge1[ii],TOE[ntoe].edge2[ii]);
*/
  }
  ++ntoe;

}                /*  for (icc=20; icc<1000; icc=icc+40) {  ... */
}



TOE[0].num = ntoe;


/* Write data for display with pgim. */
outfu = fopen_write("TOE.dat");
fprintf(outfu,"sci 3\n");
for (tt=0; tt<TOE[0].num; ++tt) {
  for (ii=0; ii<TOE[tt].nedge; ++ii) {
    fprintf(outfu," %d %d \n",TOE[tt].column,TOE[tt].edge1[ii]);
    fprintf(outfu," %d %d \n",TOE[tt].column,TOE[tt].edge2[ii]);
  }
}
fprintf(outfu,"plot\n");
fprintf(outfu,"quit\n");
fclose(outfu);


/* Fit polynomials. */

/* Order 3 : 2.46 to 1.88 (K band: 2.0-2.4) */
/* Order 4 : 1.85 to 1.42 (H band: 1.5-1.8) */
/* Order 5 : 1.48 to 1.13 (J band: 1.1-1.4) */
/* Order 6 : 1.23 to 0.95 (z/J band) */
/* Order 7 : 1.06 to 0.92 (z band:~0.82-1.03) */
/*                        (i band:~0.69-0.84) */

/* orders #3 to #7. */
for (tso=3; tso<=7; ++tso) {

  nn=0; kk=7-tso;
  for (ii=0; ii<TOE[0].num; ++ii) { if (TOE[ii].column < 1000) {
    xx[nn]=TOE[ii].column; yy[nn]=TOE[ii].edge1[kk]; ww[nn]=1.0;
    ++nn;
  }}
  if (kk != 0) {
    kk=kk-1;
    for (ii=0; ii<TOE[0].num; ++ii) { if (TOE[ii].column > 1000) {
      xx[nn]=TOE[ii].column; yy[nn]=TOE[ii].edge1[kk]; ww[nn]=1.0;
      ++nn;
    }}
  }
  order = 2;
  if (GJ_polyfit(nn,xx,yy,ww,order,0,&xoffa[tso],coefa[tso]) != 1) { printf("***error:%d:fita failed.\n",tso); exit(1); }

  sprintf(wrd,"fita%d.tbl",tso);
  outfu=fopen_write(wrd);
  fprintf(outfu,"| xx    | yy    |  ff       |\n");
/*                1234567 1234567 12345678912 */
  for (ii=0; ii<nn; ++ii) {
    xv = xx[ii] - xoffa[tso];
    ff = cpolyval((order+1),coefa[tso],xv);
    fprintf(outfu," %7.1f %7.1f %12.5f \n",xx[ii],yy[ii],ff);
  }
  fclose(outfu);


  nn=0; kk=7-tso;
  for (ii=0; ii<TOE[0].num; ++ii) { if (TOE[ii].column < 1000) {
    xx[nn]=TOE[ii].column; yy[nn]=TOE[ii].edge2[kk]; ww[nn]=1.0;
    ++nn;
  }}
  if (kk != 0) {
    kk=kk-1;
    for (ii=0; ii<TOE[0].num; ++ii) { if (TOE[ii].column > 1000) {
      xx[nn]=TOE[ii].column; yy[nn]=TOE[ii].edge2[kk]; ww[nn]=1.0;
      ++nn;
    }}
  }
  order = 2;
  if (GJ_polyfit(nn,xx,yy,ww,order,0,&xoffb[tso],coefb[tso]) != 1) { printf("***error:%d:fitb failed.\n",tso); exit(1); }
  sprintf(wrd,"fitb%d.tbl",tso);
  outfu=fopen_write(wrd);
  fprintf(outfu,"| xx    | yy    |  ff       |\n");
/*                1234567 1234567 12345678912 */
  for (ii=0; ii<nn; ++ii) {
    xv = xx[ii] - xoffb[tso];
    ff = cpolyval((order+1),coefb[tso],xv);
    fprintf(outfu," %7.1f %7.1f %12.5f \n",xx[ii],yy[ii],ff);
  }
  fclose(outfu);

}


return;
}


/* ----------------------------------------------------------------------
 Find background region from a profile (also find test object region).
*/
void tsx_background_region( int tso, int npro, double tpro[9][300], short tbck[9][300], short tobj[9][300] )
{
/**/
int narr,ii,kk;
/**/
float p95,p80,median,arr[300];
/**/
/* Clear, set up, and sort. */
for (ii=0; ii<300; ++ii) { tbck[tso][ii]=0; tobj[tso][ii]=0; }
narr=0;
for (ii=0; ii<npro; ++ii) { arr[narr]=tpro[tso][ii]; ++narr; }
median=cfind_median(narr,arr); 
kk=0.80*(narr-1); p80=arr[kk];
kk=0.95*(narr-1); p95=arr[kk];
for (ii=0; ii<npro; ++ii) {
  if (tpro[tso][ii] < p80) { tbck[tso][ii] = 1; }
  if (tpro[tso][ii] > p95) { tobj[tso][ii] = 1; }
}
return;
}


/* ----------------------------------------------------------------------
 Find background region from a profile (also find test object region).
 This assumes that this is an A-B type profile. -tab 28apr2014
*/
void tsx_background_region_mb( int tso, int npro, double tpro[9][300], short tbck[9][300], short tobj[9][300] )
{
/**/
int narr,ii,kk;
/**/
float p95,p80,p20,median,arr[300];
/**/

/* Clear, set up, and sort. */
for (ii=0; ii<300; ++ii) { tbck[tso][ii]=0; tobj[tso][ii]=0; }

/* Load. */
narr=0;
for (ii=0; ii<npro; ++ii) { arr[narr]=tpro[tso][ii]; ++narr; }
median=cfind_median(narr,arr); 

/* background. */
kk=0.20*(narr-1); p20=arr[kk];
kk=0.80*(narr-1); p80=arr[kk];
for (ii=0; ii<npro; ++ii) {
  if ((tpro[tso][ii] > p20)&&(tpro[tso][ii] < p80)) { tbck[tso][ii] = 1; }
}

/* object. */
kk=0.95*(narr-1); p95=arr[kk];
for (ii=0; ii<npro; ++ii) {
  if (tpro[tso][ii] > p95) { tobj[tso][ii] = 1; }
}

return;
}



/* ----------------------------------------------------------------------
  Find minimum width for this echelle order.
*/
int tsx_minwidth( int ecol, int tso, double polyoff[], TSPtype TSP[] )
{
/**/
int width,lower,upper,lo,ii;
/**/
lo=499;
for (ii=20; ii<ecol-20; ii=ii+50) {
  lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
  upper = tsx_find_image_row( 2, ii, tso, polyoff, TSP );
  width = 1 + upper - lower;
  if (width < lo) { lo=width; }
}
if (lo < 1) { printf("***error:a: lo<1 .\n"); exit(1); }
return(lo);
}



/* ----------------------------------------------------------------------
 Find median and RMS of a region of image (special case for tsx_image_filter).
 Avoid adjacent pixels to (ii,jj) when computing median and rms.  -tab 02jun2014
*/
void tsx_medrms( int irad, int jrad, int lower, int upper, float refimg[],
                 int nc, int ii, int jj, float *median, float *rms )
{
/**/
int kk,narr,ppp,iii,jjj,ii1,ii2,jj1,jj2;
int ii1a,ii2a,jj1a,jj2a;
float arr[808];
/**/
/* Include. */
ii1=ii-irad; ii2=ii+irad;
jj1=jj-jrad; jj2=jj+jrad;
if (ii1 <   0  ) ii1=0;
if (ii2 > nc-1 ) ii2=nc-1;
if (jj1 < lower) jj1=lower;
if (jj2 > upper) jj2=upper;
/* Avoid. */
ii1a=ii-1; ii2a=ii+1;
jj1a=jj-1; jj2a=jj+1;
if (ii1a <   0  ) ii1a=0;
if (ii2a > nc-1 ) ii2a=nc-1;
if (jj1a < lower) jj1a=lower;
if (jj2a > upper) jj2a=upper;
narr=0;
for (iii=ii1; iii<=ii2; ++iii) {
for (jjj=jj1; jjj<=jj2; ++jjj) {
  if ((iii < ii1a)||(iii > ii2a)) {
  if ((jjj < jj1a)||(jjj > jj2a)) {
    ppp = iii+(jjj*nc);
    arr[narr] = refimg[ppp];
    ++narr;
  }}
}}
if (narr < 4) { printf("***error:tmr: too few pixels for median [%d].\n",narr); exit(1); }
*median = cfind_median(narr,arr);
*rms=0.;
for (kk=0; kk<narr; ++kk) {
  *rms = *rms + ((arr[kk] - *median) * (arr[kk] - *median));
}
*rms = sqrt( *rms / (float)narr );
return;
}



/* ----------------------------------------------------------------------
  Median filter an entire image, but only replace pixels beyond a threshold.
  If rejsign=+1., only reject very large positive deviations.
  If rejsign=-1., only reject very large negative deviations.
  -tab 30may2014
*/
void tsx_image_filter( float image[], float refimg[], int nc, int nr, int irad, int jrad, float siglim, 
                       double polyoff[], TSPtype TSP[], float rejsign )
{
/**/
int count,pixno,size,ii,jj;
int tso,ecol,mnum,lower,upper;
int ii1,ii2,jj1,jj2,iii,jjj;
/**/
float median,rms,dev;
/**/
/* Check. */
size = ((irad+1+irad) * (jrad+1+jrad));
if (size > 800) { printf("***error: tmi: irad,jrad combination is too big.\n"); exit(1); }
if (size < 9)   { printf("***error: tmi: irad,jrad combination is too small.\n"); exit(1); }
/* Copy. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii+(jj*nc);
  refimg[pixno] = image[pixno];
/*  sigimg[pixno] = 0.; */
}}
/* Only look at pixels within an echelle order. */
count=0;
for (tso=3; tso<8; ++tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);
  for (ii=0; ii<ecol; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    upper = lower + mnum - 1;
    for (jj=lower; jj<=upper; ++jj) {
      pixno= ii +(jj    *nc);
      tsx_medrms( irad,jrad,lower,upper, refimg, nc,ii,jj, &median, &rms );
      dev = (image[pixno] - median) / rms;
      if ((rejsign*dev) > siglim) { 
        image[pixno] = median; 
/*      sigimg[pixno] = dev; */
        ++count; 
/* Check adjacent pixels at lower threshold. */
        ii1=ii-1; ii2=ii+1; jj1=jj-1; jj2=jj+1;
        if (ii1 <   0  ) ii1=0;
        if (ii2 > nc-1 ) ii2=nc-1;
        if (jj1 < lower) jj1=lower;
        if (jj2 > upper) jj2=upper;
        for (iii=ii1; iii<=ii2; ++iii) {
        for (jjj=jj1; jjj<=jj2; ++jjj) {
          pixno= iii +(jjj    *nc);
          tsx_medrms( irad,jrad,lower,upper, refimg, nc,iii,jjj, &median, &rms );
          dev = (image[pixno] - median) / rms;
          if ((rejsign*dev) > siglim/3.) { 
            image[pixno] = median; 
/*          sigimg[pixno] = dev; */
            ++count; 
          }
        }}
      }
    }
  }
}
printf("Filter: Replaced %d low pixels in A-B image (due to B image CRs).\n",count);
/* #@#
tsx_wout("refimg.fits",refimg,nc,nr);
tsx_wout("newimg.fits",image ,nc,nr);
tsx_wout("sigimg.fits",sigimg,nc,nr);
*/
return;
}



/* ----------------------------------------------------------------------
  Interpolate over masking.
  irad=7 and jrad=2 yields a box 15cols by 5rows ..
*/
void tsx_mask_interpolate( float image[], float maskimg[], int nc, int nr, int irad, int jrad )
{
/**/
int pixno,size,pp,kk,ii,jj,ii1,ii2,jj1,jj2,iii,jjj;
float arr4[808];
int bad,count;
/**/
size = ((irad+1+irad) * (jrad+1+jrad));
if (size > 800) { printf("***error: tmi: irad,jrad combination is too big.\n"); exit(1); }
bad=0; count=0;
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii+(jj*nc);
  if ((maskimg[pixno] > 0.001)||(maskimg[pixno] < -0.001)) {   /* hot or low pixel */
    ii1=ii-irad;  ii2=ii+irad;
    jj1=jj-jrad;  jj2=jj+jrad;
    if (ii1 < 0) ii1=0; if (ii2 > nc-1) ii2=nc-1;
    if (jj1 < 0) jj1=0; if (jj2 > nr-1) jj2=nr-1;
    kk=0;
    for (iii=ii1; iii<=ii2; ++iii) {
    for (jjj=jj1; jjj<=jj2; ++jjj) {
      pp = iii+(jjj*nc);
      if ((maskimg[pp] < 0.001)&&(maskimg[pp] > -0.001)) {   /* good pixel */
        arr4[kk] = image[pp];  
        ++kk;
      }
    }}
    if (kk < 3) { printf("***error:tmi: too few pix for mdn (%d,%d) [%d].\n",ii,jj,kk); exit(1); }
    image[pixno] = cfind_median(kk,arr4);
  }
}}
return;
}


/* ----------------------------------------------------------------------
 Stats on divided flats. 
*/
double tsx_flat_stats( float image1[], float image2[], float image3[], 
                     int nc, int nr, int sc, int ec, int sr, int er )
{
/**/
int ii,jj,pixno;
/**/
double rms,sum1,sum2,sum3,num,mean1,mean2,mean3,eperdn;
/**/

/* Check and echo. */
if ((sc < 0)||(sc > nc-1)||(ec < 0)||(ec > nc-1)) { printf("***error: columns out of range.\n"); exit(1); }
if ((sr < 0)||(sr > nr-1)||(er < 0)||(er > nr-1)) { printf("***error:    rows out of range.\n"); exit(1); }

/* Means. */
sum1=0.; sum2=0.; sum3=0.; num=0.;
for (ii=sc; ii<=ec; ++ii) {
for (jj=sr; jj<=er; ++jj) {
  pixno=ii+(jj*nc);
  sum1=sum1+image1[pixno];
  sum2=sum2+image2[pixno];
  sum3=sum3+image3[pixno];
  num =num + 1.0;
}}
if (num < 1.) { printf("***error: tfs: too few pixels.\n"); exit(1); }
mean1=sum1/num;
mean2=sum2/num;
mean3=sum3/num;

/* RMS of 3. */
sum3=0.;
for (ii=sc; ii<=ec; ++ii) {
for (jj=sr; jj<=er; ++jj) {
  pixno=ii+(jj*nc);
  sum3 = sum3 + ( (image3[pixno] - mean3) * (image3[pixno] - mean3) );
}}
rms = sqrt( sum3 / num );

eperdn = 2.0 / ( mean1 * rms * rms );

printf(" %12.4f %12.4f %12.8f %12.8f %12.6f \n",mean1,mean2,mean3,rms,eperdn);

return(eperdn);
}



/*----------------------------------------------------------------------
  Create variance image.
*/
void tsx_create_variance( float image[], float flatimg[], int nc, int nr, float var[] )
{
/**/
int ii,jj,pixno;
double raw,variance,ratio;
/**/
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  raw = image[pixno] * flatimg[pixno];
  if (raw > 1.) { variance = ((raw+RAWOFF) / EPERDN) + ROVDN; ratio = image[pixno] / raw; }
           else { variance = (   1.0       / EPERDN) + ROVDN; ratio = 1.;                 }
  var[pixno] = variance * ratio * ratio;
}}
return;
}



/*----------------------------------------------------------------------
 Divide an image by flat.   -tab 16apr2014
*/
void tsx_flat_divide( float image[], int nc, int nr, float flatimg[] )
{
/**/
int ii,jj,pixno;
/**/
/* Ignore first 250 rows and only use flat values >0 and <100. */
for (jj=250; jj<nr; ++jj) {
  for (ii=0; ii<nc; ++ii) {
    pixno = ii + (jj*nc);
    if ((flatimg[pixno] < 100.)&&(flatimg[pixno] > 0.)) {
      image[pixno] = image[pixno] / flatimg[pixno];
    }
  }
}
return;
}

/*----------------------------------------------------------------------
 Divide an dark image by flat.   -tab 16apr2014
 Only correct tso=3 (reddest echelle order) and only to a certain column.
*/
void tsx_dark_flat_divide( float image[], int nc, int nr, float flatimg[], double polyoff[], TSPtype TSP[] )
{
/**/
int ii,jj,pixno,lower,mnum,ecol,tso;
/**/
/* Only do top order. Each column, up to 1800.. flatten within echelle order. */
tso=3; ecol=nc;
mnum = tsx_minwidth(ecol,tso,polyoff,TSP);
for (ii=0; ii<1800; ++ii) {
  lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
  for (jj=lower-9; jj<(lower+mnum+5); ++jj) {
    pixno = ii+(jj*nc);
    if ((flatimg[pixno] < 100.)&&(flatimg[pixno] > 0.)) {
      image[pixno] = image[pixno] / flatimg[pixno];
    }
  }
}
return;
}



/*----------------------------------------------------------------------
 Shift object image to positive DN.        -tab 11apr2014
*/
void tsx_shift_image( float image[], int nc, int nr, float maskimg[], float scrimg[] )
{
/**/
int ii,jj,pixno;
float shift,median;
/**/
/* Copy and interpolate scratch image. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  scrimg[pixno] = image[pixno];
}}
tsx_mask_interpolate( scrimg, maskimg, nc, nr, 7,2 );
median = cfind_median((nc*nr),scrimg);
ii = cnint(( ((float)(nc*nr)) * 0.05 ));
shift = scrimg[ii];
if (VERB) printf("NOTE: shift image by %0.3f \n",shift);
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  image[pixno] = image[pixno] - shift;
}}
return;
}


/*----------------------------------------------------------------------
 Dark correction with ad hoc scaling.          -tab 11apr2014
 ( Generally do not use, since A-B does good job of getting rid of dark... )
*/
void tsx_correct_dark( float image[], int nc, int nr, float dark[], float scrimg[] )
{
/**/
int ii,jj,kk,pixno;
float bck,bck_dark,median;
/**/
/* Find background in upper left corner. */
kk=0;
for (ii=12; ii<91; ++ii) {
for (jj=925; jj<986; ++jj) {
  pixno = ii + (jj*nc);
  scrimg[kk] = image[pixno];
  ++kk;
}}
median = cfind_median(kk,scrimg);
ii = (float)kk * 0.10;
bck = scrimg[ii];
/* Same statistic in dark image. */
kk=0;
for (ii=12; ii<91; ++ii) {
for (jj=925; jj<986; ++jj) {
  pixno = ii + (jj*nc);
  scrimg[kk] = dark[pixno];
  ++kk;
}}
median = cfind_median(kk,scrimg);
ii = (float)kk * 0.10;
bck_dark = scrimg[ii];
printf("NOTE: Correct dark, scale by (img/dark): %0.3f / %0.3f = %0.3f .\n",bck,bck_dark,bck/bck_dark);
/* Scale dark and subtract off image. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  image[pixno] = image[pixno] - (dark[pixno] * (bck / bck_dark));
}}
return;
}


/* ----------------------------------------------------------------------
 For a given echelle order (tso), find the 'apix' value and relative
 arcseonds 'rarcs' value.
*/
void tsx_relative_arcsec( int tso, double mm, TOFFtype TOFF[], double *apix, double *rarcs )
{
/**/
double xv,yv;
/**/
*apix = mm;
if (tso > 3) {
  xv = mm - TOFF[tso].ps_xoff;
  yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
  *apix = *apix - yv;
}
xv = *apix - TOFF[3].ps_xoff;
*rarcs = cpolyval( TOFF[3].polyorder+1, TOFF[3].ps_coef, xv );
return;
}




/* ----------------------------------------------------------------------
 Find object window automatically.  -tab 28apr2014
*/
void tsx_autox( TOFFtype TOFF[], int npro, double tpro_mb[9][300], double bckpro_mb[9][300], 
                short tbck_mask[9][300], short tobj_mask[9][300], double *asp1, double *asp2 )
{
/**/
int ii,jj,tso,seg[300],iihi;
int nn,ii1,ii2,ii1x,ii2x,pp;
/**/
double rms,num,apix,rarcs1[9],rarcs2[9];
double fmbg[300];
/**/

/* Look at reddest 4 echelle orders. */
nn=0;
for (tso=3; tso<7; ++tso) {

/* Clear. */
  for (ii=0; ii<300; ++ii) { seg[ii]=0; fmbg[ii]=0.; }

/* Flux - background array. */
  for (ii=0; ii<npro; ++ii) { fmbg[ii] = tpro_mb[tso][ii] - bckpro_mb[tso][ii]; }

/* Find largest continuous segment of object pixels. */
  for (ii=0; ii<npro; ++ii) {
    if (tobj_mask[tso][ii]) {
      jj=ii+1;
      while ((jj < npro)&&(tobj_mask[tso][jj])) { ++jj; }
      seg[ii] = jj - ii;
    }
  }
  iihi=0;
  for (ii=0; ii<npro; ++ii) { if (seg[ii] > seg[iihi]) iihi=ii; }
  ii1=iihi; ii2=iihi+(seg[iihi]-1);

/* Find RMS in background mask pixels. */
rms=0.; num=0.;
for (ii=0; ii<npro; ++ii) {
  if (tbck_mask[tso][ii]) {
    rms = rms + (fmbg[ii] * fmbg[ii]);
    num = num + 1.;
  }
}
rms = sqrt( rms / num );

/* Go outward from object region while height is above N*RMS.. */
pp=ii1; while ((pp >=  0 )&&(fmbg[pp] > 2.*rms)) { --pp; }
ii1x=pp+1;
pp=ii2; while ((pp < npro)&&(fmbg[pp] > 2.*rms)) { ++pp; }
ii2x=pp-1;

tsx_relative_arcsec( tso, (double)ii1x, TOFF, &apix, &rarcs1[nn] );
tsx_relative_arcsec( tso, (double)ii2x, TOFF, &apix, &rarcs2[nn] );

printf("tso=%d  rms=%6.2f  object segment =%4d %3d    expanded =%4d %3d  [ %6.2f  %6.2f ]\n",tso,rms,ii1,ii2,ii1x,ii2x,rarcs1[nn],rarcs2[nn]);

++nn;

}

/* Median the 4 values. */
*asp1 = cfind_median8(nn,rarcs1);
*asp2 = cfind_median8(nn,rarcs2);

return;
}





/* ----------------------------------------------------------------------
  Apply dark median image subtraction.
  This uses pixels outside of the echelle orders to compute scale function.
*/
void OLD_tsx_dark_correction( float maskimg[], float dark[], float image[], int nc, int nr, double polyoff[], TSPtype TSP[] )
{
/**/
int row_up,row_dn,ii,ii1,jj,mnum,polyorder;
int again,pixno,tso,good,nn8,ok;
/**/
double sum1,sum2,num,xoff,coef[9],ff,xv;
double high,wrms,wppd;
/**/
const int mm=200;
double xx8[mm],yy8[mm],ww8[mm];
/**/
FILE *outfu;

/* Look at pixels outside of echelle orders (middle orders only: 4,5 . */
nn8=0;
for (tso=4; tso<=5; ++tso) {
  mnum = tsx_minwidth(nc,tso+1,polyoff,TSP);
  for (ii1=100; ii1<nc-102; ii1=ii1+100) {
    sum1=0.; sum2=0.; num=0.;
    for (ii=ii1; ii<(ii1+100); ++ii) {
      row_up = tsx_find_image_row( 1, ii, tso,   polyoff, TSP ) - 12;
      row_dn = tsx_find_image_row( 1, ii, tso+1, polyoff, TSP ) + mnum + 12;
      for (jj=row_dn; jj<=row_up; ++jj) {
        pixno = ii+(jj*nc);
        if (maskimg[pixno] < 0.001) {
          sum1 = sum1 + dark[pixno];
          sum2 = sum2 + image[pixno];
          num  = num + 1.;
        }
      }
    }
    if (num < 3.) { printf("***error:tdc: too few pixels [%f].\n",num); exit(1); }
    xx8[nn8] = sum1 / num;
    yy8[nn8] = sum2 / num;
    ww8[nn8] = 1.0;
    ++nn8;
    if (nn8 > mm-3) { printf("***error:tdc: too many pts.\n"); exit(1); }

  }
}

/* Fit polynomial. */
if (nn8 < 8) { printf("***error:tdc:too few points, nn8=%d\n",nn8); exit(1); }
polyorder=2;
again=1; good=nn8;
while (again) {
  again=0;
  ok = GJ_polyfit( nn8, xx8, yy8, ww8, polyorder, 0, &xoff, coef );
  if (!ok) { printf("***error:tdc:bad polyfit.\n"); exit(1); }
  GJ_polyfit_residuals( nn8, xx8, yy8, ww8, polyorder, xoff, coef, &high, &wrms, &wppd );
  printf("good=%4d  high=%9.4f wrms=%9.4f wppd=%9.4f\n",good,high,wrms,wppd);
  for (ii=0; ii<nn8; ++ii) { if (ww8[ii] > 0.) {
    xv=xx8[ii]-xoff; ff=cpolyval( polyorder+1, coef, xv );
    if (ABS((yy8[ii]-ff)) > (2.*wrms)) { ww8[ii]=0.; again=1; }
  }}
  good=0;
  for (ii=0; ii<nn8; ++ii) { if (ww8[ii] > 0.) ++good; }
  if (good < nn8/2) again=0;
}


/* #@# */
outfu = fopen_write("comp.tbl");
fprintf(outfu,"| xx       | yy       | ww        | ff       |\n");
for (ii=0; ii<nn8; ++ii) {
  xv=xx8[ii]-xoff; ff=cpolyval( polyorder+1, coef, xv );
  fprintf(outfu," %10.3f %10.3f %10.3f %10.3f \n",xx8[ii],yy8[ii],ww8[ii],ff);
}
fclose(outfu);
/* #@# */


/* Correct dark image on unmasked pixels. */
for (pixno=0; pixno<(nc*nr); ++pixno) {
  xv = dark[pixno] - xoff;
  ff = cpolyval( polyorder+1, coef, xv );
  if (maskimg[pixno] < 0.001) { image[pixno] = image[pixno] - ff; }
}

return;
}


/*----------------------------------------------------------------------
 tst3 :  Test Case 3 - - average and normalize flat fields...
       (read in 'tst3.ls')
*/
void tsx_tst3( int nc, int nr, float maskimg[], float scrimg[], double polyoff[], TSPtype TSP[] )
{
/**/
int mnum,tso,ecol,lower,kk,nn,ii,jj,pixno;
int hdutype,anynull;
int polyorder,ok;
int ii1,ii2,iii;
/**/
char ffile[200];
char ffiles[22][200];
char wrd[200];
/**/
long nbuffer;
long firstpixel = 1;
int status = 0;
float nullval = 0;  /* don't check for null values in the image */
/**/
float *image;
float *sumimg;
float *divimg;
float *normimg;
float ave;
/**/
double sum,num,fsp[5000];
double xx[200],yy[200],ww[200],xoff,coef[9],midrow,mean,xv;
/**/
fitsfile *fptr;
/**/
FILE  *infu;
FILE  *outfu;
/**/
/**/

/* Init. */
printf("tst3:(tst3.ls): nc=%d nr=%d\n",nc,nr);
nbuffer   = nc * nr;

/* Allocate. */
image  = (float *)calloc(((nc*nr)+1000),sizeof(float));
sumimg = (float *)calloc(((nc*nr)+1000),sizeof(float));
divimg = (float *)calloc(((nc*nr)+1000),sizeof(float));
normimg= (float *)calloc(((nc*nr)+1000),sizeof(float));

/* Load filenames. */
nn=0;
infu = fopen_read("tst3.ls");
while (fgetline(ffiles[nn],infu)) { 
  if (nn >= 20) { printf("***error: too many files.\n"); exit(1); }
  ++nn; 
}
fclose(infu);
printf("Found %d files.\n",nn);

/* Clear. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc); sumimg[pixno] = 0.;
}}

/* Read files (no shift). */
for (kk=0; kk<nn; ++kk) {
  strcpy(ffile,ffiles[kk]);
  printf("Open and read: '%s'.\n",ffile);
  fits_open_file( &fptr, ffile, READONLY, &status );
  fits_movabs_hdu( fptr, 1, &hdutype, &status );
  fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image, &anynull, &status ); cfua_error(status);
  fits_close_file( fptr, &status ); cfua_error(status);
  for (ii=0; ii<nc; ++ii) {
  for (jj=0; jj<nr; ++jj) {
    pixno = ii+(jj*nc);
    sumimg[pixno] = sumimg[pixno] + image[pixno];
  }}
}

/* Average. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc);
  sumimg[pixno] = sumimg[pixno] / (float)nn;
}}
tsx_wout("flat_average.fits",sumimg,nc,nr);

/* Div. images. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc);
  divimg[pixno] = sumimg[pixno];
  normimg[pixno]= sumimg[pixno];
}}

/* . . . create flat spectra. . . */
for (tso=7; tso>=3; --tso) {

  for (ii=0; ii<nc; ++ii) { fsp[ii]=0.; }
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);

/* For each column, fit the rows to create a spectrum. */
  for (ii=0; ii<ecol; ++ii) {

/* Load values. */
    kk=0;
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower+9; jj<(lower+mnum-9); ++jj) {
      pixno = ii+(jj*nc);
      if (maskimg[pixno] < 0.001) {
        xx[kk] = (double)jj;
        yy[kk] = sumimg[pixno];
        ww[kk] = 1.;
        ++kk;
      }
    }
    if (kk < 20) { printf("***error: too few pts.. kk=%d\n",kk); exit(1); }

/* Spectrum at middle row. */
    midrow = ( (double)(lower+mnum) + (double)lower) / 2.;
    polyorder=2;
    ok = GJ_polyfit( kk, xx, yy, ww, polyorder, 0, &xoff, coef );
    if (!ok) { printf("***error:tst3:bad polyfit.\n"); exit(1); }
    xv = midrow - xoff;
    fsp[ii] = cpolyval( polyorder+1, coef, xv );

/* #@# Divide fit to create normalized flat.
    for (jj=lower-10; jj<(lower+mnum+10); ++jj) {
      pixno = ii+(jj*nc);
      xv = (double)jj - xoff;
      fv = cpolyval( polyorder+1, coef, xv );
      normimg[pixno] = normimg[pixno] / fv;
    }
*/

/* Divide flat by mean in each column in a given echelle order. */
    ii1 = ii - 0;   ii2 = ii + 0;
    if (ii1 < 0) ii1=0;
    if (ii2 > (nc-1)) ii2=nc-1;
    mean=0.; num=0.;
    for (iii=ii1; iii<=ii2; ++iii) {
    for (jj=lower+5; jj<(lower+mnum-5); ++jj) {
      pixno = iii+(jj*nc);
      if (maskimg[pixno] < 0.001) {
        mean = mean + sumimg[pixno];
        num  = num + 1.;
      }
    }
    }
    if (num > 0.) mean = mean / num;
    for (jj=lower-10; jj<(lower+mnum+10); ++jj) {
      pixno = iii+(jj*nc);
      if (mean > 0.) { normimg[pixno] = normimg[pixno] / mean; }
                else { normimg[pixno] = 0.; }
    }

  }     /* for (ii=0; ii<ecol; ++ii) { ... */

/* Write spectrum, find average. */
  sprintf(wrd,"fsp%d.tbl",tso);
  outfu = fopen_write(wrd);
  fprintf(outfu,"| col |  fsp       |\n");
  sum=0.; num=0.;
  for (ii=0; ii<nc; ++ii) { 
    fprintf(outfu," %5d %12.3f \n",ii,fsp[ii]); 
    sum=sum+fsp[ii];
    num=num+1.;
  }
  fclose(outfu);
  ave = sum / num;

/* Divide each pixel in echelle order by 'ave'. */
  for (ii=0; ii<ecol; ++ii) {
    num=0.; sum=0.;
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower-20; jj<(lower+mnum+20); ++jj) {
    if ((jj > 0.)&&(jj < nr)) {
      pixno = ii+(jj*nc);
      if (ave > 0.) { divimg[pixno] = divimg[pixno] / ave; }
               else { divimg[pixno] = 0.; }
    }}
  }

}
tsx_wout("flat_div.fits" ,divimg,nc,nr);
tsx_wout("flat_norm_bymean.fits",normimg,nc,nr);

printf("NOTE:  COPY  flat_norm_bymean.fits  to  c/tsx/cal/tsx_flat.fits to use later.\n");

/* #@# Divide flat average by flat norm.
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc);
  sumimg[pixno] = sumimg[pixno] / normimg[pixno];
}}
tsx_wout("flat_avenorm_bymean.fits",sumimg,nc,nr);
*/


/* Free. */
free(image);
free(sumimg);
free(divimg);
free(normimg);
return;
}



/*----------------------------------------------------------------------
 tst2 :  Test Case 2 - - create median dark from list of darks 
         in 'tst2.ls' and create mask flag image..
*/
void tsx_tst2( int nc, int nr, float last_maskimg[], float scrimg[], double polyoff[], TSPtype TSP[], float flatimg[] )
{
/**/
int nn,kk,ppp;
int ii,jj,pixno;
int iii,jjj,ii1,ii2,jj1,jj2;
int hdutype,anynull;
/**/
char ffile[200];
char ffiles[12][200];
/**/
long nbuffer;
long firstpixel = 1;
int status = 0;
float nullval = 0;  /* don't check for null values in the image */
/**/
double sigs,variance;
/**/
float median,arr4[900];
float *images[11];
float *mdnimg;
float *maskimg;
float *sigimg;
/**/
fitsfile *fptr;
/**/
FILE  *infu;
/**/

/* Init. */
printf("tst2:(tst2.ls): nc=%d nr=%d\n",nc,nr);
nbuffer   = nc * nr;

/* Allocate. */
for (ii=0; ii<11; ++ii) {
  images[ii] = (float *)calloc(((nc*nr)+1000),sizeof(float));
}
mdnimg = (float *)calloc(((nc*nr)+1000),sizeof(float));
maskimg= (float *)calloc(((nc*nr)+1000),sizeof(float));
sigimg = (float *)calloc(((nc*nr)+1000),sizeof(float));

/* Load filenames. */
nn=0;
infu = fopen_read("tst2.ls");
while (fgetline(ffiles[nn],infu)) { 
  if (nn >= 10) { printf("***error: too many files.\n"); exit(1); }
  ++nn; 
}
fclose(infu);
printf("Found %d files.\n",nn);

/* Read files, shift and divide each image by normalized flat. */
for (ii=0; ii<nn; ++ii) {
  strcpy(ffile,ffiles[ii]);
  printf("Open and read: '%s'.\n",ffile);
  fits_open_file( &fptr, ffile, READONLY, &status );
  fits_movabs_hdu( fptr, 1, &hdutype, &status );
  fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, images[ii], &anynull, &status ); cfua_error(status);
  tsx_shift_image( images[ii], nc, nr, last_maskimg, scrimg );
  tsx_dark_flat_divide( images[ii], nc, nr, flatimg, polyoff, TSP );
  fits_close_file( fptr, &status ); cfua_error(status);
}

/* Create median image. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii+(jj*nc);
  for (kk=0; kk<nn; ++kk) {
    arr4[kk]=images[kk][pixno];
  }
  mdnimg[pixno] = cfind_median(nn,arr4);
}
}
tsx_wout("dark_median.fits",mdnimg,nc,nr);


/* Create mask image. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii+(jj*nc);
  maskimg[pixno]=0.;
  sigimg[pixno]=0.;
}}
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
pixno = ii+(jj*nc);
if (mdnimg[pixno] > 0.) {

/* Median of nearby region. */
  ii1=ii-7; ii2=ii+7;
  jj1=jj-3; jj2=jj+3;
  if (ii1 < 0) ii1=0; if (ii2 > nc-1) ii2=nc-1;
  if (jj1 < 0) jj1=0; if (jj2 > nr-1) jj2=nr-1;
  kk=0;
  for (iii=ii1; iii<=ii2; ++iii) {
  for (jjj=jj1; jjj<=jj2; ++jjj) {
    ppp = iii+(jjj*nc);
    arr4[kk]=mdnimg[ppp];
    ++kk;
  }}
  median = cfind_median(kk,arr4);
  if (median > 1.) { variance = (median / EPERDN) + ROVDN; }
              else { variance =                     ROVDN; }
  sigs = (mdnimg[pixno] - median) / sqrt(variance);
  sigimg[pixno] = sigs;

/* Mask? */
  if (sigs > 16.) {
    maskimg[pixno] = sigs;
/* Look at neighbors, mask at lower threshold. */
    ii1=ii-1; ii2=ii+1;
    jj1=jj-1; jj2=jj+1;
    if (ii1 < 0) ii1=0; if (ii2 > nc-1) ii2=nc-1;
    if (jj1 < 0) jj1=0; if (jj2 > nr-1) jj2=nr-1;
    for (iii=ii1; iii<=ii2; ++iii) {
    for (jjj=jj1; jjj<=jj2; ++jjj) {
      ppp = iii+(jjj*nc);
      if (mdnimg[ppp] > 0.) {
        sigs = (mdnimg[ppp] - median) / sqrt(variance);
        if (sigs > 4.) { maskimg[ppp] = sigs; }
      }
    }}
  }

}}}
tsx_wout("maskimg.fits",maskimg,nc,nr);
tsx_wout("sigimg.fits",sigimg,nc,nr);




/* #@#
/x Interpolate median image. x/
tsx_mask_interpolate( mdnimg, maskimg, nc, nr, 7,2 );
tsx_wout("tst2_mdn2.fits",mdnimg,nc,nr);

}
*/

/* #@# Higher limits for echelle order = 3.
  if (jj > 900) { Limit1=500.; Limit2=100.; } 
           else { Limit1=50.;  Limit2=20.;  }
*/


/* Free. */
for (ii=0; ii<nn; ++ii) { free(images[ii]); }
free(mdnimg);
free(maskimg);
free(sigimg);
return;
}





/*----------------------------------------------------------------------
 tst4 :  Test Case 4 - - subtract bias from median to get read-noise..
*/
void tsx_tst4( char arg[202][200], float maskimg[] )
{
/**/
int ii,jj,nc,nr,pixno;
int iii,jjj,ii1,ii2,jj1,jj2;
int hdutype,anynull;
/**/
char ffile[200];
/**/
long nbuffer;
long firstpixel = 1;
int status = 0;
float nullval = 0;  /* don't check for null values in the image */
/**/
double sum1,num,mean,rms;
/**/
float *image1, *image2, *image3;
/**/
fitsfile *fptr;
/**/
FILE  *outfu;
/**/

/* Init. */
nc=2048; nr=1024; nbuffer   = nc * nr;
image1 = (float *)calloc(((nc*nr)+1000),sizeof(float));
image2 = (float *)calloc(((nc*nr)+1000),sizeof(float));
image3 = (float *)calloc(((nc*nr)+1000),sizeof(float));

/* Argument 1. */
strcpy(ffile,arg[1]);
printf("Reading: %s ..\n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image1, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_mask_interpolate( image1, maskimg, nc, nr, 7,2 );
tsx_wout("image1.fits",image1,nc,nr);

/* Argument 2. */
strcpy(ffile,arg[2]);
printf("Reading: %s ..\n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image2, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_mask_interpolate( image2, maskimg, nc, nr, 7,2 );
tsx_wout("image2.fits",image2,nc,nr);

/* Subtract. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  image3[pixno] = image1[pixno] - image2[pixno];
}}
tsx_wout("diff.fits",image3,nc,nr);

/* Stats. */
outfu=fopen_write("stats.tbl");
fprintf(outfu,"| col  | row  | mean       | rms        |  num  |\n");
for (ii=100; ii<nc-100; ii=ii+100) {
for (jj=100; jj<nr-100; jj=jj+100) {
  ii1=ii-10; ii2=ii+10;
  jj1=jj+10; jj2=jj+10;
  sum1=0.; num=0.;
  for (iii=ii1; iii<=ii2; ++iii) {
  for (jjj=jj1; jjj<=jj2; ++jjj) {
    pixno = iii + (jjj*nc);
    if (maskimg[pixno] < 0.001) {
      sum1 = sum1 + image3[pixno];
      num  = num + 1;
    }
  }}
  mean = sum1 / num;
  rms=0.;
  for (iii=ii1; iii<=ii2; ++iii) {
  for (jjj=jj1; jjj<=jj2; ++jjj) {
    pixno = iii + (jjj*nc);
    if (maskimg[pixno] < 0.001) {
      rms  = rms + ((image3[pixno] - mean) * (image3[pixno] - mean));
    }
  }}
  rms = sqrt( rms / num);
  fprintf(outfu,"%6d %6d %12.5f %12.5f %7.1f \n",ii,jj,mean,rms,num);
}}
fclose(outfu);

return;
}






/*----------------------------------------------------------------------
 tst5 :  Test Case 5 - - divide two flats for eperdn...
*/
void tsx_tst5( char arg[202][200], float maskimg[] )
{
/**/
int kk,ii,jj,nc,nr,pixno;
int hdutype,anynull;
/**/
char ffile[200];
/**/
long nbuffer;
long firstpixel = 1;
int status = 0;
float nullval = 0;  /* don't check for null values in the image */
/**/
double rr1,rr2;
/**/
float *image1, *image2, *image3;
/**/
fitsfile *fptr;
/**/

/* Init. */
nc=2048; nr=1024; nbuffer   = nc * nr;
image1 = (float *)calloc(((nc*nr)+1000),sizeof(float));
image2 = (float *)calloc(((nc*nr)+1000),sizeof(float));
image3 = (float *)calloc(((nc*nr)+1000),sizeof(float));

/* Argument 1. */
strcpy(ffile,arg[1]);
printf("Reading: %s ..\n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image1, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_mask_interpolate( image1, maskimg, nc, nr, 7,2 );
tsx_wout("image1.fits",image1,nc,nr);

/* Argument 2. */
strcpy(ffile,arg[2]);
printf("Reading: %s ..\n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image2, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_mask_interpolate( image2, maskimg, nc, nr, 7,2 );
tsx_wout("image2.fits",image2,nc,nr);

/* Divide. */
kk=0;
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  if (image2[pixno] > 0.) { image3[pixno] = image1[pixno] / image2[pixno]; }
                     else { image3[pixno] = 0.; ++kk; }
}}
tsx_wout("div.fits",image3,nc,nr);
printf("Wrote div.fits with %d 'zero denominator' pixels.\n",kk);

/* Stats. */
rr2 = 0.;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 340, 463,925,996); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1319,1508,932,986); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 432, 531,768,817); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1184,1274,772,830); rr2 = rr2 + rr1;

rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 603, 688,633,686); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1342,1432,617,678); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 477, 625,494,551); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1013,1148,473,518); rr2 = rr2 + rr1;

rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 333, 420,342,404); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr, 532, 576,330,396); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1357,1412,439,508); rr2 = rr2 + rr1;
rr1 = tsx_flat_stats(image1,image2,image3,nc,nr,1525,1556,429,490); rr2 = rr2 + rr1;

rr2 = rr2 / 12.;
printf(" rr2 = %12.5f \n",rr2);

return;
}


/*----------------------------------------------------------------------
 tst7 :  Test Case 7 - - low pixels in flat..
*/
void tsx_tst7( char arg[202][200], float maskimg[], float flatimg[], float scrimg[], double polyoff[], 
               TSPtype TSP[], int nc, int nr )
{
/**/
int pixno,ii,jj,lower,tso,ecol,mnum;
FILE *outfu;
FILE *outfu2;
/**/
/* Clear. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii+(jj*nc);
  scrimg[pixno] = 0.;
}}
/* Look along orders.. */
outfu = fopen_write("masklow_pixels.tbl");
fprintf(outfu ,"| col | row | value   |\n");
outfu2= fopen_write("maskhigh_pixels.tbl");
fprintf(outfu2,"| col | row | value   |\n");
for (tso=3; tso<8; ++tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
  for (ii=0; ii<ecol; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      if (maskimg[pixno] > 0.001) {
        fprintf(outfu2,"%5d %5d %9.5f\n",ii,jj,maskimg[pixno]);
        scrimg[pixno] = maskimg[pixno];
        if (flatimg[pixno] < 0.24) {
          printf("this hot pixel also a low flat pixel: %5d %5d %9.5f %9.5f \n",ii,jj,maskimg[pixno],flatimg[pixno]);
        }
      } else {
        if (flatimg[pixno] < 0.24) {
          fprintf(outfu,"%5d %5d %9.5f\n",ii,jj,flatimg[pixno]);
          if (flatimg[pixno] < 0.001) {
            flatimg[pixno]=0.001;
            printf("...correcting small pixel %f  col=%d  row=%d  \n",flatimg[pixno],ii,jj);
          }
          scrimg[pixno] = -1. * flatimg[pixno];
        }
      }
    }
  }
}
tsx_wout("maskimg.fits",scrimg,nc,nr);
fclose(outfu);
fclose(outfu2);

return;
}



/*----------------------------------------------------------------------
 tst6 :  Test Case 6 - - do A and B image, subtract, analyze diff. 
*/
void tsx_tst6( char arg[202][200], float maskimg[], float flatimg[], float scrimg[] )
{
/**/
int nn8,ii,jj,jj1,jj2,ii1,ii2,skyrow[99],skycol[99],nc,nr,pixno;
int polyorder,kk,hdutype,anynull;
/**/
char ffile[200];
char wrd[200];
/**/
long nbuffer;
long firstpixel = 1;
int status = 0;
float nullval = 0;  /* don't check for null values in the image */
/**/
double sum,vsum,xoff,coef[9],xv,ff,xx8[900],yy8[900],ww8[900],gg8[900],dd8[900],ee8[900];
double num,merr,mean,rms,rr;
double mean_ratio,num_mean_ratio;
/**/
float *image1, *image2, *image3, *var1, *var2, *var3, *err3;
/**/
fitsfile *fptr;
/**/
FILE *outfu;
FILE *outfu2;
/**/

/* Init. */
nc=2048; nr=1024; nbuffer   = nc * nr;
image1= (float *)calloc(((nc*nr)+1000),sizeof(float));
image2= (float *)calloc(((nc*nr)+1000),sizeof(float));
image3= (float *)calloc(((nc*nr)+1000),sizeof(float));
var1  = (float *)calloc(((nc*nr)+1000),sizeof(float));
var2  = (float *)calloc(((nc*nr)+1000),sizeof(float));
var3  = (float *)calloc(((nc*nr)+1000),sizeof(float));
err3  = (float *)calloc(((nc*nr)+1000),sizeof(float));

/* Argument 1. */
strcpy(ffile,arg[1]);
if (VERB) printf("Reading: %s .. \n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image1, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_shift_image( image1, nc, nr, maskimg, scrimg );
tsx_flat_divide( image1, nc, nr, flatimg );
tsx_mask_interpolate( image1, maskimg, nc, nr, 3,3 );
tsx_create_variance( image1, flatimg, nc, nr, var1 );
tsx_wout("image1.fits",image1,nc,nr);
tsx_wout("var1.fits"  ,var1  ,nc,nr);


mean_ratio=0.;
num_mean_ratio=0.;
outfu2 = fopen_write("rms.tbl");
fprintf(outfu2,"|kk| rms    | merr   | ratio  | sum    |\n");
/*               12 12345678 12345678 12345678 12345678 */

skycol[0] =1184; skyrow[0] =800; 
skycol[1] = 824; skyrow[1] =661; 
skycol[2] =1441; skyrow[2] =642; 
skycol[3] =1765; skyrow[3] =628; 
skycol[4] = 410; skyrow[4] =369; 
skycol[5] = 184; skyrow[5] =664; 
skycol[6] =1301; skyrow[6] =482; 
skycol[7] =  65; skyrow[7] =533; 
skycol[8] = 590; skyrow[8] =522; 
skycol[9] =  37; skyrow[9] =535; 
skycol[10]=1475; skyrow[10]=960; 
skycol[11]=1456; skyrow[11]=641; 
for (kk=0; kk<12; ++kk) {
  jj1=skyrow[kk]-30; jj2=skyrow[kk]+30;
  ii1=skycol[kk]-3;  ii2=skycol[kk]+3;
  nn8=0;
  for (jj=jj1; jj<=jj2; ++jj) {
    sum=0.; vsum=0.;
    for (ii=ii1; ii<=ii2; ++ii) {
      pixno = ii + (jj*nc);
      sum =sum + image1[pixno];
      vsum=vsum+ var1[pixno];
    }
    xx8[nn8] = (double)jj;
    yy8[nn8] = sum;
    ww8[nn8] = 1.0;
    ee8[nn8] = sqrt(vsum);
    ++nn8;
  }
  polyorder=2;
  GJ_polyfit( nn8, xx8, yy8, ww8, polyorder, 0, &xoff, coef );
  sprintf(wrd,"sky%2.2d.tbl",kk);
  outfu = fopen_write(wrd);
  fprintf(outfu,"| row   | sum    | fit    | diff   | error  |\n");
/*                1234567 12345678 12345678 12345678 12345678 */
  for (ii=0; ii<nn8; ++ii) {
    xv = xx8[ii] - xoff;
    ff = cpolyval( polyorder+1, coef, xv );
    gg8[ii] = yy8[ii]-ff;
    fprintf(outfu," %7.1f %8.3f %8.3f %8.3f %8.3f \n",xx8[ii],yy8[ii],ff,yy8[ii]-ff,ee8[ii]);
  }
  fclose(outfu);

  sum=0.;
  for (ii=0; ii<nn8; ++ii) { sum=sum+yy8[ii]; }
  sum = sum / (double)nn8;

  merr=0.;
  for (ii=0; ii<nn8; ++ii) { merr=merr+ee8[ii]; }
  merr = merr / (double)nn8;

/* Mean and RMS. */
  mean=0.;
  for (ii=0; ii<nn8; ++ii) { mean=mean+gg8[ii]; }
  mean = mean / (double)nn8;
  rms=0.;
  for (ii=0; ii<nn8; ++ii) { 
    rr = gg8[ii] - mean;
    rms = rms + (rr * rr);
  }
  rms = sqrt( rms / (double)nn8 );

/* Deviations. */
  for (ii=0; ii<nn8; ++ii) { 
    dd8[ii] = (gg8[ii] - mean) / rms; 
  }

/* Again, without deviants. */
  mean=0.; num=0.;
  for (ii=0; ii<nn8; ++ii) { 
    if (ABS((dd8[ii])) < 3.) {
      mean=mean+gg8[ii]; 
      num =num + 1.;
    } else {
      if (VERB) printf("reject deviant: %f\n",dd8[ii]);
    }
  }
  mean = mean / num;
  rms=0.;
  for (ii=0; ii<nn8; ++ii) { 
    if (ABS((dd8[ii])) < 3.) {
      rr = gg8[ii] - mean;
      rms = rms + (rr * rr);
    }
  }
  rms = sqrt( rms / num );

/*
  printf("rms=%8.3f  merr=%8.3f  [%8.4f]\n",rms,merr,rms/merr);
  printf("\n");
*/

  fprintf(outfu2," %2d %8.3f %8.3f %8.4f %8.1f \n",kk,rms,merr,rms/merr,sum);

  if ( (rms/merr) < 2.0 ) {
    mean_ratio = mean_ratio + (rms/merr);
    num_mean_ratio = num_mean_ratio + 1.0;
  }

}
fclose(outfu2);

printf("::: %8.3f %8.3f %8.3f %8.3f \n",(mean_ratio / num_mean_ratio),EPERDN,ROVDN,RAWOFF);

/* #@#
/x Argument 2. x/
strcpy(ffile,arg[2]);
printf("Reading: %s .. \n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image2, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_shift_image( image2, nc, nr, maskimg, scrimg );
tsx_flat_divide( image2, nc, nr, flatimg );
tsx_mask_interpolate( image2, maskimg, nc, nr, 3,3 );
tsx_create_variance( image2, flatimg, nc, nr, var2 );
tsx_wout("image2.fits",image2,nc,nr);
tsx_wout("var2.fits"  ,var2  ,nc,nr);

/x Diff. x/
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj*nc);
  image3[pixno] = image1[pixno] - image2[pixno];
  var3[pixno]   = var1[pixno] + var2[pixno];
  err3[pixno]   = sqrt(var3[pixno]);
}}
tsx_wout("image3.fits",image3,nc,nr);
tsx_wout("var3.fits"  ,var3  ,nc,nr);
tsx_wout("err3.fits"  ,err3  ,nc,nr);

*/

return;
}



/*----------------------------------------------------------------------
  Compute offsets of a point source between different echelle orders.
*/
void tsx_echelle_offset()
{
/**/
double tsra[20];
double tsdec[20];
double arr8[2000];
double hip,bck,mdn,rr1,rr2,rr3,ra_deg,dec_deg;
double sum,wsum,cent;
double cents[20][8];
double ff,xv,xx8[900],yy8[900],zz8[900];
double sep,ps_xoff,ps_coef[9];
/**/
char wrd[200];
char line[200];
/**/
int polyorder,ok,nn8,ii,iihip,nn,tso,kk;
int status = 0;
/**/
FILE *infu;
FILE *outfu;
/**/
fitsfile *fptr;
/**/
float xx[2000],yy[2000];
/**/

for (kk=117; kk<125; ++kk) {

  tsra[kk-117] = 0.1;
  tsdec[kk-117]= 0.2;
  sprintf(wrd,"Tspec_20130824_%4.4d.fits",kk);
  fits_open_file( &fptr, wrd, READONLY, &status ); cfua_error(status); 
  cfua_chead(fptr,"RA" , wrd );
  if (clc(wrd) > 3) {
    rr1=GLV(wrd,1); rr2=GLV(wrd,2); rr3=GLV(wrd,3);
    ra_deg = misc_hms2deci( (int)rr1, (int)rr2, rr3, 1);
  } else { ra_deg = 0.; }
  cfua_chead(fptr,"DEC", wrd );
  if (clc(wrd) > 3) {
    rr1=GLV(wrd,1); rr2=GLV(wrd,2); rr3=GLV(wrd,3);
    dec_deg= misc_hms2deci( (int)rr1, (int)rr2, rr3, 0);
  } else { dec_deg = 0.; }
  fits_close_file( fptr, &status ); cfua_error(status); 
  tsra[kk-117] = ra_deg;
  tsdec[kk-117]= dec_deg;

  for (tso=3; tso<8; ++tso) {
  
    sprintf(wrd,"Tspec_20130824_%4.4d-pro%d.tbl",kk,tso);
    infu = fopen_read(wrd);
    nn=0;
    while (fgetline(line,infu)) { if (line[0] != '|') {
      xx[nn] = GLV(line,1);
      yy[nn] = GLV(line,2);
      arr8[nn]= yy[nn];
      ++nn;
    }}
    mdn = cfind_median8(nn,arr8);
    ii = cnint((float)nn * 0.8);
    bck = arr8[ii];
    for (ii=0; ii<nn; ++ii) { yy[ii]=yy[ii]-bck; }

/* Centroid. */
    hip=0.; iihip=0;
    for (ii=0; ii<nn; ++ii) { if (yy[ii] > hip) { hip=yy[ii]; iihip=ii; } }
    if ((iihip < 6)||(iihip > nn-6)) {  printf("***error: too close to edge\n"); exit(1); }
    sum=0.; wsum=0.;
    for (ii=(iihip-4); ii<=(iihip+4); ++ii) {
      if (yy[ii] > 0.) {
        sum = sum + (xx[ii]*yy[ii]);
        wsum= wsum+ (yy[ii]);
      }
    }
    if (wsum > 0.) { cent = sum/wsum; } else { cent=0.; }
    printf(" %4d %2d %9.3f \n",kk,tso,cent);
    cents[kk-117][tso]=cent;

/* Save. */
    sprintf(wrd,"pt%d-%d.tbl",kk,tso);
    outfu=fopen_write(wrd);
    fprintf(outfu,"| xx   | yy      |\n");
    for (ii=0; ii<nn; ++ii) { fprintf(outfu," %6.1f %9.3f \n",xx[ii],yy[ii]); }
    fclose(outfu);

  }
}


nn8=0;
for (kk=117; kk<125; ++kk) {
  sep = 3600. * cangsep( tsra[0], tsdec[0], tsra[kk-117], tsdec[kk-117] );
  sep = 10.813 - sep;    /* 10.813 is the approximate center of 3rd order slit */
  printf(" %9.5f %9.5f %9.5f %9.5f \n",cents[kk-117][3],tsra[kk-117],tsdec[kk-117],sep);
  xx8[nn8] = cents[kk-117][3];
  yy8[nn8] = sep;
  zz8[nn8] = 1.0;
  ++nn8;
}
polyorder=2;
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, &ps_xoff, ps_coef );
if (!ok) { printf("***error:bad polyfit.\n"); exit(1); }
outfu = fopen_write("arcoff.poly");
fprintf(outfu," %15.8e %15.8e %15.8e %15.8e \n",ps_xoff,ps_coef[0],ps_coef[1],ps_coef[2]);
fclose(outfu);
outfu = fopen_write("arcoff.tbl");
fprintf(outfu,"| cent | sep   | fit   |\n");
for (kk=0; kk<nn8; ++kk) {
  xv = xx8[kk] - ps_xoff;
  ff = cpolyval( polyorder+1, ps_coef, xv );
  fprintf(outfu," %6.1f %7.3f %7.3f \n", xx8[kk],yy8[kk],ff);
}
fclose(outfu);
xv = 50. - ps_xoff;
ff = cpolyval( polyorder+1, ps_coef, xv );
printf("... cent=50  ff=%9.3f\n",ff);


/* 4 - 3 */
nn8=0;
for (kk=117; kk<125; ++kk) {
  xx8[nn8] = cents[kk-117][3];
  yy8[nn8] = cents[kk-117][4] - cents[kk-117][3];
  zz8[nn8] = 1.0;
  ++nn8;
}
polyorder=2;
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, &ps_xoff, ps_coef );
if (!ok) { printf("***error:bad polyfit.\n"); exit(1); }
outfu = fopen_write("toff4m3.poly");
fprintf(outfu," %15.8e %15.8e %15.8e %15.8e \n",ps_xoff,ps_coef[0],ps_coef[1],ps_coef[2]);
fclose(outfu);
outfu = fopen_write("tst4m3.tbl");
fprintf(outfu,"| kk | cent3   | c4m3    | fit     |\n");
for (kk=0; kk<nn8; ++kk) {
  xv = xx8[kk] - ps_xoff;
  ff = cpolyval( polyorder+1, ps_coef, xv );
  fprintf(outfu," %4d %9.3f %9.3f %9.3f \n", kk+117, xx8[kk], yy8[kk], ff );
}
fclose(outfu);

/* 5 - 3 */
nn8=0;
for (kk=117; kk<125; ++kk) {
  xx8[nn8] = cents[kk-117][3];
  yy8[nn8] = cents[kk-117][5] - cents[kk-117][3];
  zz8[nn8] = 1.0;
  ++nn8;
}
polyorder=2;
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, &ps_xoff, ps_coef );
if (!ok) { printf("***error:bad polyfit.\n"); exit(1); }
outfu = fopen_write("toff5m3.poly");
fprintf(outfu," %15.8e %15.8e %15.8e %15.8e \n",ps_xoff,ps_coef[0],ps_coef[1],ps_coef[2]);
fclose(outfu);
outfu = fopen_write("tst5m3.tbl");
fprintf(outfu,"| kk | cent3   | c5m3    | fit     |\n");
for (kk=0; kk<nn8; ++kk) {
  xv = xx8[kk] - ps_xoff;
  ff = cpolyval( polyorder+1, ps_coef, xv );
  fprintf(outfu," %4d %9.3f %9.3f %9.3f \n", kk+117, xx8[kk], yy8[kk], ff );
}
fclose(outfu);

/* 6 - 3 */
nn8=0;
for (kk=117; kk<125; ++kk) {
  xx8[nn8] = cents[kk-117][3];
  yy8[nn8] = cents[kk-117][6] - cents[kk-117][3];
  zz8[nn8] = 1.0;
  ++nn8;
}
polyorder=2;
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, &ps_xoff, ps_coef );
if (!ok) { printf("***error:bad polyfit.\n"); exit(1); }
outfu = fopen_write("toff6m3.poly");
fprintf(outfu," %15.8e %15.8e %15.8e %15.8e \n",ps_xoff,ps_coef[0],ps_coef[1],ps_coef[2]);
fclose(outfu);
outfu = fopen_write("m3.tbl");
fprintf(outfu,"| kk | cent3   | c6m3    | fit     |\n");
for (kk=0; kk<nn8; ++kk) {
  xv = xx8[kk] - ps_xoff;
  ff = cpolyval( polyorder+1, ps_coef, xv );
  fprintf(outfu," %4d %9.3f %9.3f %9.3f \n", kk+117, xx8[kk], yy8[kk], ff );
}
fclose(outfu);

/* 7 - 3 */
nn8=0;
for (kk=117; kk<125; ++kk) {
  xx8[nn8] = cents[kk-117][3];
  yy8[nn8] = cents[kk-117][7] - cents[kk-117][3];
  zz8[nn8] = 1.0;
  ++nn8;
}
polyorder=2;
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, &ps_xoff, ps_coef );
if (!ok) { printf("***error:bad polyfit.\n"); exit(1); }
outfu = fopen_write("toff7m3.poly");
fprintf(outfu," %15.8e %15.8e %15.8e %15.8e \n",ps_xoff,ps_coef[0],ps_coef[1],ps_coef[2]);
fclose(outfu);
outfu = fopen_write("tst7m3.tbl");
fprintf(outfu,"| kk | cent3   | c7m3    | fit     |\n");
for (kk=0; kk<nn8; ++kk) {
  xv = xx8[kk] - ps_xoff;
  ff = cpolyval( polyorder+1, ps_coef, xv );
  fprintf(outfu," %4d %9.3f %9.3f %9.3f \n", kk+117, xx8[kk], yy8[kk], ff );
}
fclose(outfu);

return;
}



/* ----------------------------------------------------------------------
 Nightsky.  Centroid pixel position of known night sky lines.
*/
void tsx_nightsky( int tso, double tspbck_nob[9][3000], double xpos, double *cent, double *high, double *bck )
{
/**/
int kk,ii,ii1,ii2,nn,narr;
double wgt,sum,wsum,median,arr[40],xx8[40],yy8[40];
/* FILE *outfu; */
/**/
ii1 = cnint(xpos) - 9;
ii2 = cnint(xpos) + 9;
/* outfu = fopen_write("ns.dat"); */
/* fprintf(outfu,"| ii  | bck_nob |\n"); */
nn=0;
for (ii=ii1; ii<=ii2; ++ii) {
  xx8[nn]=(double)ii;
  yy8[nn]=tspbck_nob[tso][ii];
  ++nn;
/*  fprintf(outfu," %5d %9.3f \n",ii,tspbck_nob[tso][ii]); */
}
/* fclose(outfu); */
/* Centroid. */
narr=0; for (ii=0; ii<nn; ++ii) { arr[narr]=yy8[ii]; ++narr; }
median = cfind_median8(narr,arr);
kk = cnint( 0.1 * (double)narr );
*bck = arr[kk];
sum=0.; wsum=0.; *high=0.;
for (ii=6; ii<=12; ++ii) { 
  wgt = yy8[ii] - *bck;
  sum = sum + (xx8[ii] * wgt);
  wsum= wsum+ wgt;
  if (wgt > *high) *high=wgt;
}
if (wsum > 0.) { *cent = sum / wsum; } else { *cent=0.; }
return;
}


/* ----------------------------------------------------------------------
 Given x,y arrays and a new x point (xpt) return an interpolated value for
 the y axis.  x,y array must be monotonically increasing in x.  "0" is
 returned for xpt values outside of the range.
 Inputs:  n (number of points in x,y arrays).   x[],y[] (arrays).
          xpt (value to interpolate towards).
*/
/*@@*/
double tsx_yinterp(int n, double x[], double y[], double xpt)
{
/**/
  int ii;
  double yy;
/**/
  if (n   <   1   ) return(0.);
  if (xpt < x[0]  ) return(0.);
  if (xpt > x[n-1]) return(0.);
  ii=1;
  while ( xpt > x[ii] ) ++ii;
  yy = y[ii-1] + ( (y[ii]-y[ii-1]) * (xpt-x[ii-1]) / (x[ii]-x[ii-1]) );
  return(yy);
}


/* ----------------------------------------------------------------------
 Given B-V, return effective stellar temperature in degrees Kelvin
 (main sequence, solar metallicity).  Data from "Bowers and Deeming" text
 book and "Lejeune, Cuisinier, and Buser: A&ASupp 125, 229 (1997)."
*/
/*@@*/
double tsx_StarTempBmV(double userbv)
{
/**/
  double bv[27]=
{
-0.35 ,
-0.31 ,
-0.16 ,
-0.100,
-0.050,
+0.000,
+0.050,
+0.100,
+0.13 ,
+0.175,
+0.260,
+0.27 ,
+0.335,
+0.42 ,
+0.425,
+0.58 ,
+0.595,
+0.70 ,
+0.760,
+0.875,
+0.89 ,
+1.095,
+1.18 ,
+1.350,
+1.45 ,
+1.63 ,
+1.80
};
  double stp[27]=
{
 40000.,
 28000.,
 15500.,
 11750.,
 10200.,
  9500.,
  8870.,
  8600.,
  8500.,
  8000.,
  7500.,
  7400.,
  7000.,
  6580.,
  6500.,
  6030.,
  6000.,
  5520.,
  5500.,
  5000.,
  4900.,
  4500.,
  4130.,
  4000.,
  3480.,
  2800.,
  2400.
};
  double rr;
/**/
if (userbv <= bv[0] ) return(40000.);
if (userbv >= bv[26]) return(2400.);
rr = tsx_yinterp(27, bv, stp, userbv);
return(rr);
}





/* ----------------------------------------------------------------------
  Find Star in Tycho and find temperature. -tab 11aug2014 
*/
int tsx_StarTemp( TYCtype TYC[], int numTYC, double sra, double sdec, double *los, double *EffTemp )
{
/**/
int ok,ii,loii;
double losep,sep,diff,BmV;
/**/
losep=0.015; loii=-1; *EffTemp=-8.;
for (ii=0; ii<numTYC; ++ii) {
  diff = ABS((sdec - TYC[ii].mean_dec));
  if (diff < 0.015) {
    sep = cangsep(sra,sdec,TYC[ii].mean_ra,TYC[ii].mean_dec);
    if (sep < losep) {
      losep = sep; 
      loii  = ii;
/* Proper motion.
      pra = TYC[ii].mean_ra  + ( (TYC[ii].pm_ra  * 14.) / 3600000. );
      pdec= TYC[ii].mean_dec + ( (TYC[ii].pm_dec * 14.) / 3600000. );
      losep2 = cangsep(sra,sdec,pra,pdec);
*/
    }
  }
}
if (loii < 0) {
  printf("Could not find Tycho match.\n");
  ok=0;
} else {
  ok=1;
  if ((TYC[loii].bt_mag > -9.)&&(TYC[loii].vt_mag > -9.)) {
    BmV = TYC[loii].bt_mag - TYC[loii].vt_mag;
    *EffTemp = tsx_StarTempBmV( BmV );
  } else {
    BmV = -9.;
    *EffTemp = -9.;
  }
  printf("Found Tycho match, sep=%9.3f  tyra=%9.5f tydec=%9.5f EffTemp=%12.5f  BmV=%8.3f.\n",
          losep*3600.,TYC[loii].mean_ra,TYC[loii].mean_dec,*EffTemp,BmV);
}
   

                    /* Tycho Star Catalog
                    #define MAXTYC 2701000
                    struct TYCstructure {
                      double mean_ra,mean_dec;
                      float pm_ra,pm_dec;
                      float bt_mag,vt_mag;
                      float bt_mag_error,vt_mag_error;
                    };
                    typedef struct TYCstructure TYCtype;
                    */


*los = losep;
return(ok);
}



/* ----------------------------------------------------------------------
  Sky fit with rejections.  -tab 30may2014
*/
int tsx_skyfit( int nn8, double xx8[], double yy8[], double zz8[], double vv8[], int polyorder,
                double *bck_xoff, double bck_coef[], double siglim, int tso, int icol )
{
/**/
int totrej,nrej,iter,ii,ok;
double xv,ff,rms,num,dev;
/**/
ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, bck_xoff, bck_coef );
if (siglim < 99.) {
  totrej=0; iter=0; nrej=99;
  while ((ok)&&(totrej < nn8/4)&&(iter < 4)&&(nrej > 0)) {
    rms=0.; num=0.;
    for (ii=0; ii<nn8; ++ii) { if (zz8[ii] > 0.) {
      xv = xx8[ii] - *bck_xoff;
      ff = cpolyval( polyorder+1, bck_coef, xv );
      rms = rms + ((yy8[ii] - ff) *(yy8[ii] - ff));
      num = num + 1.;
    }}
    rms = sqrt(( rms / num ));
    nrej=0;
    for (ii=0; ii<nn8; ++ii) { if (zz8[ii] > 0.) {
      xv = xx8[ii] - *bck_xoff;
      ff = cpolyval( polyorder+1, bck_coef, xv );
      dev = ABS((yy8[ii] - ff));
      if (dev > rms*siglim) { zz8[ii]=0.; ++nrej; ++totrej; }
    }}
    ok = GJ_polyfit( nn8, xx8, yy8, zz8, polyorder, 0, bck_xoff, bck_coef );
    ++iter;
  }
}

/* #@# was 205 ..
if ((icol == 248)&&(tso == 6)) {
  printf("#@# writing tt.tbl test file within tsx_skyfit() .. icol=%d  tso=%d\n",icol,tso);
  outfu = fopen_write("tt.tbl");
  fprintf(outfu,"| xx      | yy      | ee      | ff      | zz      |\n");
  for (ii=0; ii<nn8; ++ii) {
    xv = xx8[ii] - *bck_xoff;
    ff = cpolyval( polyorder+1, bck_coef, xv );
    if (vv8[ii] > 0.) { ee=sqrt(vv8[ii]); } else { ee=0.; }
    fprintf(outfu," %9.5f %9.5f %9.5f %9.5f %9.5f \n",xx8[ii],yy8[ii],ee,ff,zz8[ii]);
  }
  fclose(outfu);
}
   #@# */


return(ok);
}


/* ----------------------------------------------------------------------
  Return 1 if NO absorption line.
*/
int tsx_NoAbsorption( double ww, int Van, double Va1[], double Va2[] )
{
/**/
int ii,NoA;
/**/
NoA = 1;
for (ii=0; ii<Van; ++ii) { if ((ww > Va1[ii])&&(ww < Va2[ii])) NoA=0; }
return(NoA);
}



/* ----------------------------------------------------------------------
  Vega matching..
*/
void tsx_vega( int tso, int Vn, double Vx[], double Vy[], int VCn, double VCx[], double VCy[],
               int Van, double Va1[], double Va2[], double VyD[] )
{
/**/
int ii,jj,order;
/**/
double rr,ww,mean,xmin,xmax,meanC,num;
double xv,xoff,coef[9],xx8[9000],yy8[9000],ww8[9000];
/**/
float xx4[9000],yy4[9000],ww4[9000],bb4[9000];
/**/
/* #@#
double T,a;
char wrd[200];
FILE *outfu;
*/
/**/
double mm,DIVy;
/**/

mean=0.; xmin=Vx[0]; xmax=Vx[0];
for (ii=0; ii<Vn; ++ii) { 
  mean = mean + Vy[ii];
  if (Vx[ii] < xmin) xmin=Vx[ii];
  if (Vx[ii] > xmax) xmax=Vx[ii];
}
mean = mean / (double)Vn;
xmin = xmin * 10000.;
xmax = xmax * 10000.;

meanC=0.; num=0.;
for (ii=0; ii<VCn; ++ii) { 
  if ((VCx[ii] > xmin)&&(VCx[ii] < xmax)) { meanC = meanC + VCy[ii];  num=num+1.; }
}
meanC = meanC / num;

printf("NOTE: mean=%12.5e  meanC=%12.5e  num=%f \n",mean,meanC,num);


/* Load values. */
for (ii=0; ii<Vn; ++ii) {
  ww = Vx[ii] * 10000.;
  jj = tsx_cneari( ww, VCn, VCx );
  rr = mean * VCy[jj] / meanC; 
  xx8[ii] = ww;
  yy8[ii] = rr;
  ww8[ii] = tsx_NoAbsorption( ww, Van, Va1, Va2 );
  xx4[ii] = (float)xx8[ii];
  yy4[ii] = (float)yy8[ii];
  ww4[ii] = (float)ww8[ii];
}


/* Fit Black Body. */
/*
tsx_FitBlackBody( Vn, xx4, yy4, ww4, &T, &a );
for (ii=0; ii<Vn; ++ii) { 
  bb4[ii] = tsx_BlackBodyAT( a, T, (double)xx4[ii] ); 
}
*/

/* Fit polynomial. */
order=4;
if (tso == 6) order=5;
if (GJ_polyfit(Vn,xx8,yy8,ww8,order,0,&xoff,coef) != 1) { printf("***error:tv:polyfit failed.\n"); exit(1); }
for (ii=0; ii<Vn; ++ii) { 
  xv = xx8[ii] - xoff;
  bb4[ii] = (float)cpolyval((order+1),coef,xv);
}







/* Note: Need to smooth out 'mm' array below... */
/* Note: Need to smooth out 'mm' array below... */
/* Note: Need to smooth out 'mm' array below... */
/* Note: Need to smooth out 'mm' array below... */
/* Note: Need to smooth out 'mm' array below... */

/* #@# 
sprintf(wrd,"V_%d.tbl",tso);   printf("Write '%s'.\n",wrd);
outfu = fopen_write(wrd);
fprintf(outfu,"| wave       | sum        | vega       |  bb        | ww4    | mm         |  NEWy      |  DIVy      |\n");
*/
for (ii=0; ii<Vn; ++ii) {
  ww = Vx[ii] * 10000.;
  jj = tsx_cneari( ww, VCn, VCx );
  rr = mean * VCy[jj] / meanC; 
  mm = rr / (double)bb4[ii];
  VyD[ii] = Vy[ii] / mm;
  DIVy = Vy[ii] / rr;
/* #@#
  fprintf(outfu," %12.5e %12.5e %12.5e %12.5e %8.1f %12.5e %12.5e %12.5e \n",
          Vx[ii],Vy[ii],rr,bb4[ii],ww4[ii]*35000.,mm,VyD[ii],DIVy);
*/
}
/* #@#
fclose(outfu);
  #@# */


return;
}




/* ----------------------------------------------------------------------
 Load vega calibration spectrum.    -tab 13aug2014
*/
void tsx_load_vega( char tsxdir[], int *VCn, double VCx[], double VCy[], double VCb[],
                    int *Van, double Va1[], double Va2[] )
{
/**/
FILE *infu;
char wrd[200];
char line[200];
double xx,sum,num,wave;
double xxos[30000],yyos[30000],yyoss[30000];
int iii,ii1,ii2,jj,kk,ii,nn;
FILE *outfu;
/**/
float xx4[30000],yy4[30000],ww4[30000];
double T,a;
/**/

/* Read absorption data. */
nn=0;
sprintf(wrd,"%scal/vega_abs.dat",tsxdir);
infu = fopen_read(wrd);
while (fgetline(line,infu)) { Va1[nn]=GLV(line,1); Va2[nn]=GLV(line,2); ++nn; }
fclose(infu);
*Van = nn;

/* Read data. */
nn=0;
sprintf(wrd,"%scal/VegaCal.tbl",tsxdir);
infu = fopen_read(wrd);
while (fgetline(line,infu)) { if (line[0] != '|') {
  wave = GLV(line,1);
  if ((wave > 8900.)&&(wave < 27000.)) {
    VCx[nn] = wave;
    VCy[nn] = GLV(line,2);
    ++nn;
  }
}}
fclose(infu);

/* Sample at 1 Angstrom. */
kk=0;
for (xx=9000.; xx<26000.; xx=xx+1.) {
  jj = tsx_cneari(xx,nn,VCx);
  xxos[kk] = xx;
  yyos[kk] = VCy[jj];
  ++kk;
}

/* Smooth. */
jj=0;
for (ii=0; ii<kk; ++ii) {
  ii1 = ii - 7;
  ii2 = ii + 7;
  if (ii1 < 0   ) ii1=0;
  if (ii2 > kk-1) ii2=kk-1;
  sum=0.; num=0.;
  for (iii=ii1; iii<ii2; ++iii) { sum=sum+yyos[iii]; num=num+1.; }
  sum = sum / num;
  yyoss[ii] = sum;
  VCx[jj] = xxos[ii];
  VCy[jj] = yyoss[ii];
  ++jj;
}
*VCn = jj;


if (VERB) {
outfu = fopen_write("v2.tbl");
fprintf(outfu,"| xxos       | yyos       | yyoss      |\n");
for (ii=0; ii<kk; ++ii) { fprintf(outfu," %12.5e %12.5e %12.5e \n",xxos[ii],yyos[ii]*1.e+10,yyoss[ii]*1.e+10); }
fclose(outfu);
}

for (ii=0; ii<*VCn; ++ii) { 
  xx4[ii] = VCx[ii];
  yy4[ii] = VCy[ii];
  ww4[ii] = 1.0;
}
nn = *VCn;

tsx_FitBlackBody( nn, xx4, yy4, ww4, &T, &a );
for (ii=0; ii<*VCn; ++ii) { VCb[ii] = tsx_BlackBodyAT( a, T, VCx[ii] ); }



if (VERB) {
outfu = fopen_write("v1.tbl");
fprintf(outfu,"| VCx        | VCy        | VCb        |\n");
for (ii=0; ii<*VCn; ++ii) { fprintf(outfu," %12.5e %12.5e %12.5e \n",VCx[ii],VCy[ii]*1.e+10,VCb[ii]*1.e+10); }
fclose(outfu);
}




/*
            Given a spectrum, estimate the black-body temperature "T" and the scale
            parameter "a", assuming the flux is in ergs/sec/cm^2/Angstrom
            (i.e. energy per unit wavelength):
                 flux(a,T,x) = a * ( (h*c*c)/x^5 ) / ( exp((h*c)/(k*T*x)) - 1 )
            Input: nn    : number of points in spectrum.
                   xx[]  : wavelengths in Angstroms.
                   yy[]  : flux in ergs/sec/cm^2/Angstroms.
                   ww[]  : weight values.
           Output: T     : Temperature in Kelvin.
                   a     : Temperature in Kelvin.

           void tsx_FitBlackBody( int nn, float xx[], float yy[], float ww[],
                                 double *T, double *a )
*/


printf("End of tsx_load_vega..\n");
return;
}




/* ----------------------------------------------------------------------
 Find Hot star for telluric correction. Read data which can then be applied
 to this spectrum.    -tab 13aug2014
 Returns '1' if found appropriate hot star for telluric correction.
 Vega is Teff=9600 (approx)..
*/
int tsx_find_hotstar( char tsxout[], char root[], TLStype TLS[], char HOTroot[] )
{
/**/
int kk,ii,vc,lc,pp,loii,lodif;
char wrd[200];
char wrd2[200];
char line[200];
double rr;
FILE *infu;
/**/
/* Find primary object. */
strcpy(HOTroot,"");
kk=-1;
for (ii=0; ii<TLS[0].num; ++ii) { if (strcmp(root,TLS[ii].root)==0) { kk=ii; } }
if (kk < 0) { printf("===warning: could not find this object in tsx.tbl file.\n"); return(0); }
/* Find appropriate hot star. */
lodif=9999999; loii=0;
for (ii=0; ii<TLS[0].num; ++ii) {  if (ii != kk) {
/* Right temperature? */
if ((TLS[ii].EffTemp > 9100.)&&(TLS[ii].EffTemp < 10100.)) {
/* Does spectrum exist? */
  sprintf(wrd,"%s%s-sp3.tbl",tsxout,TLS[ii].root);
  if (FileExist(wrd) == 1) {
/* Does it have VsumD values? */
    infu = fopen_read(wrd);
    vc=0; lc=0; pp=0;
    while ((fgetline(line,infu))&&(vc < 10)&&(lc < 100)) {
      if (line[0]=='|') { pp = cindex(line,"VsumD"); } else {
        if (pp > 0) {
          substrcpy_terminate(line,pp-1,clc(line),wrd2,0); rr=GLV(wrd2,1); if (rr > 0.) ++vc;
        }
      }
      ++lc;
    }
    fclose(infu);
    if (vc > 9) {
      if (ABS((kk-ii)) < lodif) { lodif= ABS((kk-ii)); loii = ii; }
    }
  }
}
}}
/* Echo. */
if (lodif > 999999) {
  printf("===warning: could not find appropriate hot star.\n");
} else {
  printf("Best match for atmospheric correction star is '%s' with Teff=%f ..\n",TLS[loii].root,TLS[loii].EffTemp);
  strcpy(HOTroot,TLS[loii].root);
}
return(1);
}




/* ----------------------------------------------------------------------
 Load tsns[] values with possible shift to sky lines.
*/
void tsx_load_tsns( int *ret_tsns_n, double tsns_x[], double tsns_y[], double *ret_ns_shift,
                    int verbose, int tso, char tsxdir[], int ecol, double tspbck_nob[9][3000] )
{
/**/
char wrd[200];
/**/
FILE *infu;
FILE *outfu;
/**/
int tsns_n,kk;
/**/
double ns_shift,wrms,mrms,num,wsum,sum,xpos;
double ns_x[900],ns_cent[900],ns_high[900],ns_sig[900],ns_diff[900],ns_wgt[900],ns_bck[900];
/**/

/* Read. */
sprintf(wrd,"%scal/tsns%d.dat",tsxdir,tso);
if (FileExist(wrd) == 0) { printf("***error: cannot find sky line ID file '%s'.\n",wrd); exit(1); }
infu = fopen_read(wrd);
tsns_n=0;
while (fgetline(wrd,infu) == 1) {
  tsns_x[tsns_n] = GLV(wrd,1);
  tsns_y[tsns_n] = GLV(wrd,2);
  ++tsns_n;
}
fclose(infu);

/* Shift to night sky lines (centroid in pixels). */
sprintf(wrd,"nss_%d.tbl",tso);
sum=0.;
for (kk=0; kk<tsns_n; ++kk) {
  xpos = (double)(ecol-1) - tsns_x[kk];
  tsx_nightsky( tso, tspbck_nob, xpos, &ns_cent[kk], &ns_high[kk], &ns_bck[kk] );
  sum = sum + ns_bck[kk];
  ns_x[kk]   = (double)(ecol-1) - ns_cent[kk];
  ns_diff[kk]= ns_x[kk] - tsns_x[kk];
  ns_wgt[kk] = 1.0;
}
ns_shift=0.;

/* 1.Weighted average and RMS. */
sum=0.; wsum=0.;
for (kk=0; kk<tsns_n; ++kk) {
  sum = sum + (ns_wgt[kk] * ns_diff[kk]);
  wsum= wsum+ ns_wgt[kk];
}
if (wsum < 0.00001) { 
  printf("===warning: not enough weight on sky lines.\n");
  ns_shift=0.;
} else {
  ns_shift = sum / wsum;
  wrms=0.; num=0.;
  for (kk=0; kk<tsns_n; ++kk) {
    wrms = wrms + (ns_wgt[kk] * (ns_diff[kk] - ns_shift) * (ns_diff[kk] - ns_shift));
    if (ns_wgt[kk] > 0.) num=num+1.;
  }
  wrms = sqrt( wrms / wsum );
  mrms = sqrt( wrms / wsum ) / sqrt(num);
/* Remove bad points. */
  for (kk=0; kk<tsns_n; ++kk) {
    ns_sig[kk] = ABS(( ns_diff[kk] / wrms ));
    if (ns_sig[kk] > 3.) ns_wgt[kk]=0.;
  }
/* 2.Weighted average and RMS. */
  sum=0.; wsum=0.;
  for (kk=0; kk<tsns_n; ++kk) {
    sum = sum + (ns_wgt[kk] * ns_diff[kk]);
    wsum= wsum+ ns_wgt[kk];
  }
  if (wsum < 0.00001) { 
    printf("===warning: not enough weight on sky lines.\n");
    ns_shift = 0.;
  } else {
    ns_shift = sum / wsum;
    wrms=0.; num=0.;
    for (kk=0; kk<tsns_n; ++kk) {
      wrms = wrms + (ns_wgt[kk] * (ns_diff[kk] - ns_shift) * (ns_diff[kk] - ns_shift));
      if (ns_wgt[kk] > 0.) num=num+1.;
    }
    wrms = sqrt( wrms / wsum );
    mrms = sqrt( wrms / wsum ) / sqrt(num);
/* Remove bad points. */
    for (kk=0; kk<tsns_n; ++kk) {
      ns_sig[kk] = ABS(( ns_diff[kk] / wrms ));
      if (ns_sig[kk] > 3.) ns_wgt[kk]=0.;
    }
/* 3.Weighted average and RMS. */
    sum=0.; wsum=0.;
    for (kk=0; kk<tsns_n; ++kk) {
      sum = sum + (ns_wgt[kk] * ns_diff[kk]);
      wsum= wsum+ ns_wgt[kk];
    }
    if (wsum < 0.00001) {
      printf("===warning: not enough weight on sky lines.\n");
      ns_shift=0.;
    } else {
      ns_shift = sum / wsum;
      wrms=0.; num=0.;
      for (kk=0; kk<tsns_n; ++kk) {
        wrms = wrms + (ns_wgt[kk] * (ns_diff[kk] - ns_shift) * (ns_diff[kk] - ns_shift));
        if (ns_wgt[kk] > 0.) num=num+1.;
      }
      wrms = sqrt( wrms / wsum );
      mrms = sqrt( wrms / wsum ) / sqrt(num);
/* Shift? */
      if (ABS(ns_shift/mrms) < 10.) { ns_shift = 0.; }
    }
  }
}
/* Echo. */
if (verbose) {
if (ABS(ns_shift) > 0.) { printf("Apply night sky line shift of %9.3f pixels.\n",ns_shift); } 
                   else { printf("Do not apply night sky line shift.\n"); }
}
/* Write table. */
if (verbose) {
  outfu = fopen_write(wrd);
  fprintf(outfu,"| x       | cent    | diff  | high   | bck    | sig    |\n");
  for (kk=0; kk<tsns_n; ++kk) {
    fprintf(outfu," %9.3f %9.3f %7.3f %8.3f %8.3f %8.3f \n",
       tsns_x[kk],ns_x[kk],ns_diff[kk],ns_high[kk],ns_bck[kk],ns_sig[kk]);
  }
  fclose(outfu);
}
*ret_ns_shift = ns_shift;
*ret_tsns_n   = tsns_n;
return;
}



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Process DBSP raw data files.
*/
int main(int argc, char *argv[])
{
/**/
char arg[202][200];
char wrd[300];
char ffile[300];
char bffile[300];
char root[200];
char HOTroot[200];
char fdir[300];
char object[200];
char utshut[200];
char wrd1[100];
char wrd2[100];
char wrd3[100];
char line[200];
char tsxdir[100];
char tsxout[100];
char xtype[10];
char userflat[200];
/**/
const int maxff = 200;
char *ff_a[maxff];
char *ff_b[maxff];
/**/
/* int ii1,ii2,iii,jj1,jj2,jjj; */
int bk1[9],bk2[9],nbk,nabk;
int noplot,autob,rbi,autox,bflag,pp,pixno,col1,col2,row1,row2,jj;
int ibk1,ibk2,iabk1,iabk2,again,isp1,isp2,narg,ii,verbose;
int typflag,detflag,kk,kff,nff,nr,nc,bnr,bnc,hdutype,anynull;
int ecol,lower,upper,oo,npro,lo_mnum,mnum,mm,tso;
int echelle_offset,tst2,tst3,tst4,tst5,tst6,tst7,vega,icol,polyorder;
int WriteSpectra,bcps;
int numTYC,tsns_n,ok,nn8;
int status = 0;
int FoundAC,HSn;
/**/
float arr4[800];
float nullval = 0;  /* don't check for null values in the image */
float *bimage;
float *bimage_V;
float *bckimg_mb;
float *bckimg_nob;
float *image_mb;
float *image_mb_V;
float *image_nob;
float *image_nob_V;
float *RBimage;
float *RBimageV;
float *RBwave;
float *maskimg;
float *darkimg;
float *flatimg;
float *scrimg;
float value;
/**/
int RBrow,ii1,ii2,jj1,jj2,iii,jjj,ppp,mmm,bbb;
float rms,dev,mean;
/**/
FILE *infu;
FILE *outfu;
/**/
double offT,EffTemp,los,wgt,exsp1,exsp2,rsp1,rsp2;
double mfarr8[2000],arr8b[2000],arr8[2000];
double abk1[39],abk2[39],asp1,asp2,fmbg_nob,fmbg_mb,disp,hidisp,lodisp;
double wave,yv0,yv,xv,rarcs,apix,tsns_xoff,tsns_coef[9],xx8[900],yy8[900],zz8[900],vv8[900];
double tsns_x[900],tsns_y[900],bck_xoff,bck_coef[9];
double num,vsum,wsum,sum,averge,exptime;
double high,wrms,wppd;
double airmass,rr,rr1,rr2,rr3,ra_deg,dec_deg,ha_deg,tzd,taz;
double xoffa[9],coefa[9][9],xoffb[9],coefb[9][9];
double polyoff[9];
double mash[300];
double bmash[300];
double bckmash[300];
double mcol[300];
double bmcol[300];
double bckmcol[300];
double btpro[9][300];
double tpro_mb[9][300];
double tpro_nob[9][300];
double tpro_nob_wgt[9][300];
double tpro_nob_mf[9][300];
double tpro_fmbg_nob_mf[9][300];
double bckpro_nob[9][300];
double bckpro_mb[9][300];
double tspbck_mb[9][3000];
double tspbck_nob[9][3000];
double tspobj[9][3000];
double tsps2n[9][3000];
double tsperr[9][3000];
double lojdsep,jdsep,jd,high_fmbg_nob_mf[9],high_nob_mf[9];
double rawrow,ratio,sep;
double ns_shift;
double Vx[3000],Vy[3000],VyD[3000],Vac[3000],HSy[3000];
double VCx[30000],VCy[30000],VCb[30000];
double Va1[300],Va2[300];
/**/
int Van,VCn,Vn,tsprow[9][3000];
/**/
short tbck_mask[9][300];
short tobj_mask[9][300];
/**/
long nbuffer;
long firstpixel = 1;
/**/
fitsfile *fptr;
fitsfile *bfptr;
/**/
TSPtype TSP[9];
TOFFtype TOFF[9];
TOFFItype TOFFI[9];
TLStype *TLS;
TYCtype *TYC;
/**/
FILE *listfu;
/**/
double offy;
/*
FILE *tstfu;
*/
double nopr[9][300];
double nopr_sum;
/**/

/* Allocate. */
TLS = (TLStype *)calloc(MAXTLS,sizeof(TLStype));
TYC = (TYCtype *)calloc(MAXTYC,sizeof(TYCtype));

/* Read in tsx.tbl if it exists. */
TLS[0].num=0; 
if (FileExist("tsx.tbl")) {
  infu = fopen_read("tsx.tbl");
  kk=0;
  while (fgetline(line,infu)) { 
    if (line[0] != '|') {
      substrcpy_terminate(line,108,111,TLS[kk].xtype,0);
      substrcpy_terminate(line,113,132,wrd,0); wrd[clc(wrd)+1]='\0'; strcpy(TLS[kk].root,wrd);
      TLS[kk].ra = cvalread0(line,36,44);
      TLS[kk].dec= cvalread0(line,46,54);
      TLS[kk].air= cvalread0(line,56,60);
      TLS[kk].exp= cvalread0(line,83,88);
      TLS[kk].EffTemp = cvalread0(line,140,148);
      substrcpy_terminate(line,1,19,wrd,0);
      TLS[kk].jd = misc_zulu_to_julian(wrd);
      ++kk;
      if (kk > MAXTLS-3) { printf("***error: too many entries in 'tsx.tbl' max is %d .\n",MAXTLS); exit(1); }
    }
  }
  fclose(infu);
  TLS[0].num=kk;
}

/* #@#
for (ii=0; ii<TLS[0].num; ++ii) {
  printf(" xtype=%s  root=%s  ra=%9.5f  dec=%9.5f  jd=%f\n", TLS[ii].xtype, TLS[ii].root, TLS[ii].ra, TLS[ii].dec, TLS[ii].jd);
}
cpauseit();
   #@# */


/* Set values for polynomial offset to align profiles. */
/* Based on Tspec_20130824_0111, with airmass~1.0 . */
polyoff[0] = 0.;
polyoff[1] = 0.;
polyoff[2] = 0.;
polyoff[3] = 0.;
polyoff[4] = 1.62;
polyoff[5] = 3.54;
polyoff[6] = 4.61;
polyoff[7] = 5.68;
polyoff[8] = 0.;

/* Clear. */
for (tso=0; tso<9; ++tso) {

/* Profiles. */
  for (ii=0; ii<300; ++ii) {
    btpro[tso][ii]       =0.;
    tpro_mb[tso][ii]     =0.;    /* Profile with possible B subtraction */
    tpro_nob[tso][ii]    =0.;    /* Profile with 'no B' subtraction */
    tpro_nob_wgt[tso][ii]=0.;    /* Profile with 'no B' subtraction-- 1/bckgnd weighted */
    bckpro_mb[tso][ii]   =0.;
    bckpro_nob[tso][ii]  =0.;
    nopr[tso][ii]        =0.;
  }

/* Masks. */
  for (ii=0; ii<300; ++ii) {
    tbck_mask[tso][ii]   =0;
    tobj_mask[tso][ii]   =0;
  }

/* Clear spectra. */
  for (ii=0; ii<3000; ++ii) {
    tspbck_mb[tso][ii] =0.;
    tspbck_nob[tso][ii]=0.;
    tspobj[tso][ii]    =0.;
    tsprow[tso][ii]    =0;
    tsps2n[tso][ii]    =0.;
    tsperr[tso][ii]    =0.;
  }
}

/* Copy arguments. */
cargcopy(argv, argc, &narg, arg);

/* Syntax. */
if (narg == 0) { tsx_syntax(); exit(0); }

/* Options. */
verbose  = cfindarg(arg,&narg,"-verbose" ,'s',wrd);

/* Tests. */
echelle_offset = cfindarg(arg,&narg,"-echelle_offset",'s',wrd);
tst2= cfindarg(arg,&narg,"-tst2",'s',wrd);
tst3= cfindarg(arg,&narg,"-tst3",'s',wrd);
tst4= cfindarg(arg,&narg,"-tst4",'s',wrd);
tst5= cfindarg(arg,&narg,"-tst5",'s',wrd);
tst6= cfindarg(arg,&narg,"-tst6",'s',wrd);
tst7= cfindarg(arg,&narg,"-tst7",'s',wrd);

vega= 1;

/* Automatic object extraction. */
rbi   = cfindarg(arg,&narg,"-rbi",'s',wrd);
autox = cfindarg(arg,&narg,"-autox",'s',wrd);
autob = cfindarg(arg,&narg,"-autob",'s',wrd);
noplot= cfindarg(arg,&narg,"-noplot",'s',wrd);


/* Spectral extraction. */
exsp1=-999.; exsp2=-999.;
if (cfindarg(arg,&narg,"sp=" ,':',wrd) == 1) { exsp1=GLV(wrd,1);  exsp2=GLV(wrd,2); }

/* Spectral extraction on arcsecond scale. */
asp1=-999.; asp2=-999.;
if (cfindarg(arg,&narg,"asp=",':',wrd) == 1) { asp1=GLV(wrd,1);  asp2=GLV(wrd,2); }

/* Background range(s). */
nbk=0; again=1;
while (again) {
  if (cfindarg(arg,&narg,"bk=" ,':',wrd) == 1) {  
    bk1[nbk]=GLV(wrd,1);   bk2[nbk]=GLV(wrd,2); ++nbk;  
    again=1;
    if (nbk > 38) { printf("***error:bk=: too many background ranges given.\n"); exit(1); }
  } else { again=0; }
}
nabk=0; again=1;
while (again) {
  if (cfindarg(arg,&narg,"abk=" ,':',wrd) == 1) {  
    abk1[nabk]=GLV(wrd,1);   abk2[nabk]=GLV(wrd,2); ++nabk;  
    if (nabk > 38) { printf("***error:abk=: too many background ranges given.\n"); exit(1); }
    again=1;
  } else { again=0; }
}

/* Vega-like star for Atmospheric Correction. */
strcpy(HOTroot,"");
if (cfindarg(arg,&narg,"ac=" ,':',wrd) == 1) {  strcpy(HOTroot,wrd); }

/* User flat field. */
strcpy(userflat,"");
if (cfindarg(arg,&narg,"flat=",':',wrd) == 1) { strcpy(userflat,wrd); }

/* List file. */
if (cfindarg(arg,&narg,"list=",':',wrd) == 1) { 
  listfu = fopen_write(wrd);
  printf("Writing list of files to '%s'.\n",wrd);
} else {
  listfu = stdout;
}


/* Check TSXOUT environment variable. */
sprintf(wrd,"=%s=",getenv("TSXOUT"));
if (strcmp(wrd,"=(null)=") == 0) {
  if (VERB) printf("NOTE: No TSXOUT environment variable, writing output to default directory.\n");
  strcpy(tsxout,"");
} else {
  substrcpy_terminate(wrd,1,clc(wrd)-1,tsxout,0);
  if (FileExist(tsxout) == 0) {
    printf("***error: Cannot find tsxOUT directory '%s'.\n",tsxout); exit(1);
  }
  tsx_append_slash(tsxout);
  if (VERB) printf("NOTE: Output will be written to (TSXOUT) '%s'.\n",tsxout);
}

/* Check TSXDIR environment variable. */
sprintf(wrd,"=%s=",getenv("TSXDIR"));
if (strcmp(wrd,"=(null)=") == 0) {
  printf("***error: You must set the environment variable TSXDIR to the 'tsx' directory, e.g '/home/ptf/tsx' .\n");
  exit(1);
}

/* Get environment variable. */
strcpy(tsxdir,getenv("TSXDIR"));
if (FileExist(tsxdir) == 0) { printf("***error: The directory '%s', does not appear to exist.\n",tsxdir); exit(1); }
tsx_append_slash(tsxdir);
if (VERB) printf("NOTE: Found TSXDIR = '%s'.\n",tsxdir);

/* Calibration Info. */
tsx_load_cal( tsxdir, TSP, TOFF, TOFFI, TYC, &numTYC );


/* Read Vega data. */
VCn=0;
tsx_load_vega( tsxdir, &VCn, VCx, VCy, VCb, &Van, Va1, Va2 ); 



/* #@#
printf("VCn=%d\n",VCn);
outfu = fopen_write("VC-tst.tbl");
for (ii=0; ii<VCn; ++ii) { 
  fprintf(outfu," %12.5e %12.5e \n",VCx[ii],VCy[ii]);
}
fclose(outfu);
*/



/* Use arcsecond range? */
WriteSpectra=0;
if (asp1 > -900.) {
  xv     = asp1 - TOFFI[3].ps_xoff;
  exsp1  = (( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
  xv     = asp2 - TOFFI[3].ps_xoff;
  exsp2  = (( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
} else {
  if ((exsp1 > -1.)&&(exsp2 > exsp1)) {
    tsx_relative_arcsec( 3, exsp1, TOFF, &apix, &rarcs ); asp1 = rarcs;
    tsx_relative_arcsec( 3, exsp2, TOFF, &apix, &rarcs ); asp2 = rarcs;
  }
}
if (asp1  > -900.) { printf("tsx: asp=%6.2f %6.2f\n",asp1,asp2);               }
if (exsp2 >   -1.) { printf("tsx:  sp=%6.2f %6.2f\n",exsp1,exsp2); WriteSpectra=1; }

for (ii=0; ii<nabk; ++ii) { printf("tsx: abk=%6.2f %6.2f\n",abk1[ii],abk2[ii]); }
for (ii=0; ii<nbk;  ++ii) { printf("tsx:  bk=%6d %6d\n", bk1[ii], bk2[ii]);     }

/* Guess at range to construct background spectra for mash weighting. -tab 04apr2014 */
if ((WriteSpectra == 0)&&(exsp2 < 0)) { exsp1 = 40; exsp2 = 60; }


/* Hot and low pixel mask from darks and flats. */
sprintf(ffile,"%scal/maskimg.fits",tsxdir);
if (VERB) printf("Reading mask: %s ..\n",ffile);
nc=2048; nr=1024; nbuffer = nc * nr;
maskimg = (float *)calloc(((nc*nr)+1000),sizeof(float));
scrimg  = (float *)calloc(((nc*nr)+1000),sizeof(float));
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, maskimg, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);

/* Read medianed dark, then shift it. */
darkimg = NULL;
if (READ_DARKIMG) {
sprintf(ffile,"%scal/darkimg.fits",tsxdir);
printf("Reading dark image: %s ..\n",ffile);
nc=2048; nr=1024; nbuffer = nc * nr;
darkimg= (float *)calloc(((nc*nr)+1000),sizeof(float));
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, darkimg, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);
tsx_shift_image( darkimg, nc, nr, maskimg, scrimg );
}


/* Read in normalized flat image. */
if (strcmp(userflat,"") == 0) { sprintf(ffile,"%scal/flatimg.fits",tsxdir); } 
                         else { strcpy(ffile,userflat);                     }
printf("Reading flat image: %s ..\n",ffile);
nc=2048; nr=1024; nbuffer = nc * nr;
flatimg = (float *)calloc(((nc*nr)+1000),sizeof(float));
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }
fits_movabs_hdu( fptr, 1, &hdutype, &status );
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, flatimg, &anynull, &status ); cfua_error(status);
tsx_mask_interpolate( flatimg, maskimg, nc, nr, 3,3 );
fits_close_file( fptr, &status ); cfua_error(status);


/* Test routines. */

if (tst2) { tsx_tst2( nc, nr, maskimg, scrimg, polyoff, TSP, flatimg ); return(0); }
if (tst3) { tsx_tst3( nc, nr, maskimg, scrimg, polyoff, TSP ); return(0); }
if (tst4) { tsx_tst4(arg,maskimg); return(0); }
if (tst5) { tsx_tst5(arg,maskimg); return(0); }
if (tst6) { tsx_tst6(arg,maskimg,flatimg,scrimg); return(0); }
if (tst7) { tsx_tst7(arg,maskimg,flatimg,scrimg,polyoff,TSP,nc,nr); return(0); }
if (echelle_offset) { tsx_echelle_offset();  return(0); }


/* Header. */
if (listfu != stdout) {
fprintf(listfu,"|  UTSHUT           | OBJECT       | ra      | dec     | air | nc | nr |averge|d|t| expt  |hi3wb|hi3mb|bcps|type| root               | los | EffTemp | file_directory \n");
/*               ----+----1----+---- ----+----1---- 123456789 123456789 12345 1234 1234 123456 1 1 1234567 12345 12345 1234 1234 ----+----1----+----2 12345 123456789   */
}


/* Allocations. */
for (ii=0; ii<maxff; ++ii) { 
  ff_a[ii] = (char *)calloc(80,sizeof(char)); 
  ff_b[ii] = (char *)calloc(80,sizeof(char)); 
}


/* File list. */
if (cindex(arg[1],".ls") > 0) {
  nff=0;
  infu = fopen_read(arg[1]);
  while (fgetline(line,infu) == 1) {
    if (nff >= maxff) { printf("***error: too many files, limit is %d .\n",maxff); exit(1); }
    strcpy(ff_a[nff],line);
    strcpy(ff_b[nff],"");
    ++nff;
  }
} else {
  if (narg == 1) {
    strcpy(ff_a[0],arg[1]);
    strcpy(ff_b[0],"");
  } else {
  if (narg == 2) {
    strcpy(ff_a[0],arg[1]);
    strcpy(ff_b[0],arg[2]);
  }}
  nff=1;
}
printf("There are %d FITS file(s) (pairs).\n",nff);


/* - - - - - - BIG LOOP - - - - - - - */
for (kff=0; kff<nff; ++kff) { printf("\n");
  
/* Construct filename, open FITS file. */
strcpy(ffile,ff_a[kff]);
jj=cindex(ffile,".fits"); 
if (jj == -1) { strcat(ffile,".fits"); jj=cindex(ffile,".fits"); }
ii=cindex_reverse(ffile,"/");  
substrcpy_terminate(ffile,ii+1,jj-1,root,0);
if (ii < 1) { strcpy(fdir,""); } else { substrcpy_terminate(ffile,0,ii-1,fdir,0); }
status=0;
printf("Open FITS file '%s'.\n",ffile);
fits_open_file( &fptr, ffile, READONLY, &status );
if (status != 0) { printf("***error: problem reading '%s'.\n",ffile); exit(1); }

/* B image? */
strcpy(bffile,ff_b[kff]); 
jj=cindex(bffile,".fits"); 
if (jj == -1) { strcat(bffile,".fits"); }
if (FileExist(bffile)) { bflag=1; } else { bflag=0; strcpy(bffile,""); }

/* Echo. */
printf("Process: %4d : %s  %s \n",kff,ffile,bffile);

/* Check image. */
fits_movabs_hdu( fptr, 1, &hdutype, &status );
nc = cfua_inhead(fptr,"NAXIS1");
nr = cfua_inhead(fptr,"NAXIS2");
if ((nc < 1800)||(nr < 900)) {
  printf("===NOTE: '%s' is not a valid spectral image.\n",ffile);
  exit(1);
}

/* Info. */
cfua_chead(fptr,"UTSHUT"  ,utshut  );  
jd = misc_zulu_to_julian(utshut); 
cfua_chead(fptr,"OBJECT"  ,object  );  
cfua_chead(fptr,"RA" , wrd );
if (clc(wrd) > 3) {
  rr1=GLV(wrd,1); rr2=GLV(wrd,2); rr3=GLV(wrd,3);
  ra_deg = misc_hms2deci( (int)rr1, (int)rr2, rr3, 1);
} else { ra_deg = 0.; }
cfua_chead(fptr,"DEC", wrd );
if (clc(wrd) > 3) {
  rr1=GLV(wrd,1); rr2=GLV(wrd,2); rr3=GLV(wrd,3);
  dec_deg= misc_hms2deci( (int)rr1, (int)rr2, rr3, 0);
} else { dec_deg = 0.; }
airmass = cfua_fhead(fptr,"AIRMASS");
exptime = cfua_fhead(fptr,"EXPTIME");

/* Star temperature. */
if (tsx_StarTemp( TYC, numTYC, ra_deg, dec_deg, &los, &EffTemp ) == 0) { los=-0.0025; EffTemp=-9.; }


/* Type? */
strcpy(xtype,"unkn");
for (ii=0; ii<TLS[0].num; ++ii) {
  if (cindex(ffile,TLS[ii].root) > -1) {
    strcpy(xtype,TLS[ii].xtype);
  }
}
printf("Exposure type (from tsx.tbl) for '%s' is '%s'.\n",ffile,xtype);

/* Check B image. */
if (bflag) {
  status=0;
  printf("Open B image '%s'.\n",bffile);
  fits_open_file( &bfptr, bffile, READONLY, &status );
  if (status != 0) { printf("***error: problem reading b file '%s'.\n",bffile); exit(1); }
  fits_movabs_hdu( bfptr, 1, &hdutype, &status );
  bnc = cfua_inhead(bfptr,"NAXIS1");
  bnr = cfua_inhead(bfptr,"NAXIS2");
  if ((bnc != nc)||(bnr != nr)) { printf("===NOTE: b file '%s' does not match a file.\n",bffile); exit(1); }
}

/* Warning. */
if ((autob)&&(strcmp(xtype,"objt") == 0)&&(TLS[0].num == 0)&&(bflag == 0)) {
  printf("***error: the -autob option requires a 'tsx.tbl' file.\n");
  exit(1);
}

/* Auto-select B file? */
if ((autob)&&(strcmp(xtype,"objt") == 0)&&(TLS[0].num > 0)&&(bflag == 0)) {
  strcpy(bffile,"");
  lojdsep = (3. * (20. + exptime)) / 3600. / 24.;
  for (ii=0; ii<TLS[0].num; ++ii) { 
  if (strcmp(TLS[ii].xtype,"objt") == 0) {
  if (TLS[ii].exp > 0.1) {
  jdsep = jd - TLS[ii].jd;  if (jdsep < 0.) jdsep=-1.*jdsep;
  if (jdsep < lojdsep) {
    ratio = exptime / TLS[ii].exp;
    if ((ratio > 0.9)&&(ratio < 1.1)) {
      sep = cangsep(ra_deg,dec_deg,TLS[ii].ra,TLS[ii].dec) * 3600.;
      if ((sep < 30.)&&(sep > 5.)) { 
        sprintf(bffile,"%s.fits",TLS[ii].root); 
        lojdsep = jdsep;
      }
    }
  }}}}
/* Relax constraints. */
  if (strcmp(bffile,"") == 0) {
    lojdsep = (30. * (20. + exptime)) / 3600. / 24.;
    for (ii=0; ii<TLS[0].num; ++ii) {
    if (strcmp(TLS[ii].xtype,"objt") == 0) {
    if (TLS[ii].exp > 0.1) {
    jdsep = jd - TLS[ii].jd;  if (jdsep < 0.) jdsep=-1.*jdsep;
    if (jdsep < lojdsep) {
      ratio = exptime / TLS[ii].exp;
      if ((ratio > 0.1)&&(ratio < 10.)) {
        sep = cangsep(ra_deg,dec_deg,TLS[ii].ra,TLS[ii].dec) * 3600.;
        if ((sep < 30.)&&(sep > 5.)) { 
          sprintf(bffile,"%s.fits",TLS[ii].root); 
          lojdsep = jdsep;
        }
      }
    }}}}
  }
/* B image? */
  if (FileExist(bffile)) { bflag=1; } else { bflag=0; strcpy(bffile,""); }
  if (bflag) {
    status=0;
    printf("Open auto-selected b file '%s'.\n",bffile);
    fits_open_file( &bfptr, bffile, READONLY, &status );
    if (status != 0) { printf("***error: problem reading b file '%s'.\n",bffile); exit(1); }
    fits_movabs_hdu( bfptr, 1, &hdutype, &status );
    bnc = cfua_inhead(bfptr,"NAXIS1");
    bnr = cfua_inhead(bfptr,"NAXIS2");
    if ((bnc != nc)||(bnr != nr)) { printf("===NOTE: auto selected b file '%s' does not match a file.\n",bffile); exit(1); }
  }
}




/* Read image. */
image_nob   = (float *)calloc(((nc*nr)+1000),sizeof(float));
image_nob_V = (float *)calloc(((nc*nr)+1000),sizeof(float));
image_mb    = (float *)calloc(((nc*nr)+1000),sizeof(float));
image_mb_V  = (float *)calloc(((nc*nr)+1000),sizeof(float));
bimage      = (float *)calloc(((nc*nr)+1000),sizeof(float));
bimage_V    = (float *)calloc(((nc*nr)+1000),sizeof(float));
bckimg_nob  = (float *)calloc(((nc*nr)+1000),sizeof(float));
bckimg_mb   = (float *)calloc(((nc*nr)+1000),sizeof(float));
RBimage     = (float *)calloc(((nc*nr)+1000),sizeof(float));
RBimageV    = (float *)calloc(((nc*nr)+1000),sizeof(float));
RBwave      = (float *)calloc(((nc*10)+1000),sizeof(float));
nbuffer     = nc * nr;
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image_nob, &anynull, &status ); cfua_error(status);
fits_close_file( fptr, &status ); cfua_error(status);

/* Clear Re-Bin image. */
for (ii=0; ii<nc; ++ii) {
  for (jj=0; jj<nr; ++jj) {
    pp = ii + (jj * nc);
    RBimage[pp]=-9000.;
    RBimageV[pp]=-9000.;
  }
}

/* Shift, Flatten, Interpolate, Variance. */
printf("Flatten, Interpolate, Variance.\n");
tsx_shift_image( image_nob, nc, nr, maskimg, scrimg );
tsx_flat_divide( image_nob, nc, nr, flatimg );
tsx_mask_interpolate( image_nob, maskimg, nc, nr, 3,3 );
tsx_create_variance( image_nob, flatimg, nc, nr, image_nob_V );

/* Read b image. */
if (bflag) {
  fits_read_img( bfptr, TFLOAT, firstpixel, nbuffer, &nullval, bimage, &anynull, &status ); cfua_error(status);
  fits_close_file( bfptr, &status ); cfua_error(status);
  tsx_shift_image( bimage, nc, nr, maskimg, scrimg );
  tsx_flat_divide( bimage, nc, nr, flatimg );
  tsx_mask_interpolate( bimage, maskimg, nc, nr, 3,3 );
  tsx_create_variance( bimage, flatimg, nc, nr, bimage_V );
/* printf("Filtering B image..\n"); */
/* Not using since I need to preserve the small hot pixels for A-B subtraction. -tab 02jun2014 */
/* tsx_image_filter( bimage, scrimg, nc, nr, 0, 9, 20.0, polyoff, TSP, 1.0 );        */
}

/* Clear background image. */
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc); 
  bckimg_nob[pixno]= 0.;
  bckimg_mb[pixno] = 0.;
}}

/* Stats. */
col1 = cnint( (float)nc * 0.25 ); col2 = cnint( (float)nc * 0.75 );
row1 = cnint( (float)nr * 0.25 ); row2 = cnint( (float)nr * 0.75 );
sum=0.; num=0.;
for (ii=col1; ii<=col2; ++ii) {
  for (jj=row1; jj<=row2; ++jj) {
    pixno = ii + (jj*nc);
    sum = sum + image_nob[pixno];
    num = num + 1.;
  }
}
averge = sum / num;
if (averge > 999999.) averge = 9999.;
if (averge < -99999.) averge = -999.;


/* Find edges. */
if (EDGEFIND == 1) {
  printf("Find edges, record polys.\n");
  tsx_edge_finder(image_nob,nc,nr,xoffa,xoffb,coefa,coefb);
  outfu = fopen_write("tsx_edge.dat");
  for (kk=3; kk<=7; ++kk) {
    fprintf(outfu," %2d %15.8e %15.8e %15.8e %15.8e \n",kk,xoffa[kk],coefa[kk][0],coefa[kk][1],coefa[kk][2]);
    fprintf(outfu," %2d %15.8e %15.8e %15.8e %15.8e \n",kk,xoffb[kk],coefb[kk][0],coefb[kk][1],coefb[kk][2]);
  }
  fclose(outfu);
}


/* Default. */
ha_deg=-9.;
tzd   =-9.;
taz   =-9.;
typflag=0;
detflag=0;


/* Initial. */
lo_mnum=999999;

/* Create profile by mashing orders using 'no B' subtraction image. */
for (tso=7; tso>=3; --tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (mm=0; mm<300; ++mm) { mash[mm]=0.; mcol[mm]=0.; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
  for (ii=20; ii<ecol-20; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      if ((maskimg[pixno] < 0.001)&&(maskimg[pixno] > -0.001)) {
        mm = jj-lower;
        mash[mm] = mash[mm] + image_nob[pixno];
        mcol[mm] = mcol[mm] + 1.0;
      }
    }
  }
/* Average and store. */
  for (mm=0; mm<mnum; ++mm) { 
    if (mcol[mm] > 0.) { tpro_nob[tso][mm] = mash[mm] / mcol[mm]; }
                  else { tpro_nob[tso][mm] = 0.;                  }
  }
  if (mnum < lo_mnum) lo_mnum=mnum;
}
npro = lo_mnum;


/* Mash b image orders. */
if (bflag) {
  for (tso=7; tso>=3; --tso) {
    if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
    for (mm=0; mm<300; ++mm) { bmash[mm]=0.; bmcol[mm]=0.; }
    mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
    for (ii=20; ii<ecol-20; ++ii) {
      lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
      for (jj=lower; jj<(lower+mnum); ++jj) {
        pixno = ii+(jj*nc);
        mm = jj-lower;
        bmash[mm] = bmash[mm] + bimage[pixno];
        bmcol[mm] = bmcol[mm] + 1.0;
      }
    }
/* Average and store. */
    for (mm=0; mm<mnum; ++mm) { btpro[tso][mm] = bmash[mm] / bmcol[mm]; }
    if (mnum < lo_mnum) lo_mnum=mnum;
  }
  if (lo_mnum != npro) { printf("***error: b image npro does not match a image.\n"); exit(1); }
}


/* Find background regions (in the 'no B' subtraction image). */
for (tso=3; tso<=7; ++tso) {
  tsx_background_region( tso, npro, tpro_nob, tbck_mask, tobj_mask );
}


/* Subtract B image from primary (or just make a copy).*/
if (bflag) {
  for (ii=0; ii<nc; ++ii) {
  for (jj=0; jj<nr; ++jj) {
    pixno=ii+(jj*nc);
    value = bimage[pixno];
    image_mb[pixno] = image_nob[pixno] - value;
    image_mb_V[pixno] = image_nob_V[pixno] + bimage_V[pixno];
  }}
} else {
  for (ii=0; ii<nc; ++ii) {
  for (jj=0; jj<nr; ++jj) {
    pixno=ii+(jj*nc);
    image_mb[pixno] = image_nob[pixno];
    image_mb_V[pixno] = image_nob_V[pixno];
  }}
}



/* Create profile by mashing orders using image (with possible B subtraction). */
for (tso=7; tso>=3; --tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (mm=0; mm<300; ++mm) { mash[mm]=0.; mcol[mm]=0.; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
  for (ii=20; ii<ecol-20; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      if ((maskimg[pixno] < 0.001)&&(maskimg[pixno] > -0.001)) {
        mm = jj-lower;
        mash[mm] = mash[mm] + image_mb[pixno];
        mcol[mm] = mcol[mm] + 1.0;
      }
    }
  }
/* Average and store. */
  for (mm=0; mm<mnum; ++mm) { 
    if (mcol[mm] > 0.) { tpro_mb[tso][mm] = mash[mm] / mcol[mm]; }
                  else { tpro_mb[tso][mm] = 0.;                  }
  }
  if (mnum < lo_mnum) lo_mnum=mnum;
}


/* Find masking in the A-B profile.  This overrides previous masking. */
if (bflag) {
  for (tso=3; tso<=7; ++tso) {
    tsx_background_region_mb( tso, npro, tpro_mb, tbck_mask, tobj_mask );
  }
}



/* User input background ranges. */
if ((nbk > 0)||(nabk > 0)) {
  for (tso=3; tso<=7; ++tso) {
    for (mm=0; mm<300; ++mm) { tbck_mask[tso][mm]=0; }     /* clear. */
    mnum = tsx_minwidth(ecol,tso,polyoff,TSP);             /* find minimum width. */
/* Pixels on echelle order 3.. */
    for (ii=0; ii<nbk; ++ii) { 
      if (tso > 3) {
/* Shift pixels of echelle order 3 to another echelle order. */
        xv = (double)bk1[ii] - TOFF[tso].ps_xoff;
        yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
        ibk1 = cnint(( (double)bk1[ii] + yv ));
        xv = (double)bk2[ii] - TOFF[tso].ps_xoff;
        yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
        ibk2 = cnint(( (double)bk2[ii] + yv ));
      } else {
        ibk1=bk1[ii]; ibk2=bk2[ii];
      }
      for (mm=0; mm<mnum; ++mm) { if ((mm >= ibk1)&&(mm <= ibk2)) tbck_mask[tso][mm]=1; }
    }
/* Relative arcseconds.. */
    for (ii=0; ii<nabk; ++ii) { 
/* Convert relative arcseconds to pixels on echelle order 3. */
      xv    = abk1[ii] - TOFFI[3].ps_xoff;
      iabk1 = cnint(( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
      xv    = abk2[ii] - TOFFI[3].ps_xoff;
      iabk2 = cnint(( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
      if (tso > 3) {
/* Shift pixels of echelle order 3 to another echelle order. */
        xv = (double)iabk1 - TOFF[tso].ps_xoff;
        yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
        ibk1 = cnint(( (double)iabk1 + yv ));
        xv = (double)iabk2 - TOFF[tso].ps_xoff;
        yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
        ibk2 = cnint(( (double)iabk2 + yv ));
      } else {
        ibk1=iabk1; ibk2=iabk2;
      }
      for (mm=0; mm<mnum; ++mm) { if ((mm >= ibk1)&&(mm <= ibk2)) tbck_mask[tso][mm]=1; }
    }
  }
}


/* Eliminate large negative pixels caused by CRs in the B image using tsx_image_filter.. */
if (bflag) {
  tsx_image_filter( image_mb, scrimg, nc, nr, 3, 3, 12.0, polyoff, TSP, -1.0 );
}

/* Create image of background fit and spectra of background fit. */
for (tso=3; tso<=7; ++tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (ii=0; ii<ecol; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    upper = tsx_find_image_row( 2, ii, tso, polyoff, TSP );
    nn8=0;
    for (jj=lower; jj<=upper; ++jj) {
      mm=jj-lower;
      if (tbck_mask[tso][mm] == 1) {
        pixno=ii+(jj*nc);
        xx8[nn8] = (double)mm;
        yy8[nn8] = image_mb[pixno];
        zz8[nn8] = 1.0;
        if ((maskimg[pixno] > 0.001)||(maskimg[pixno] < -0.001)) { zz8[nn8]=0.; }
        vv8[nn8] = image_mb_V[pixno];
        ++nn8;
      }
    }
    if (nn8 < 5) { printf("***error: spec.of.bck: nn8 < 5, nn8=%d .\n",nn8); exit(1); }

/* Sky fit (with B subtraction) */
    polyorder = 3;
    ok = tsx_skyfit( nn8, xx8, yy8, zz8, vv8, polyorder, &bck_xoff, bck_coef, 5.0, tso, ii );
    if (ok == 0) { printf("***error: fitting background polynomial.\n"); exit(1); }
    xv = ((exsp1+exsp2) / 2.) - bck_xoff;
    yv = cpolyval( polyorder+1, bck_coef, xv );
    tspbck_mb[tso][ii] = yv;
    for (jj=lower; jj<=upper; ++jj) {
      mm=jj-lower;
      xv = (double)mm - bck_xoff;
      yv = cpolyval( polyorder+1, bck_coef, xv );
      pixno=ii+(jj*nc);
      bckimg_mb[pixno] = yv; 
    }
  }
}



/* #@# smooth?
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  ii1=(ii)-1; ii2=(ii)+1; jj1=(jj)-1; jj2=(jj)+1;
  if (ii1 <   0) ii1=0;
  if (ii2 >= nc) ii2=nc-1;
  if (jj1 <   0) jj1=0;
  if (jj2 >= nr) jj2=nr-1;
  sum=0.; num=0.;
  for (iii=ii1; iii<=ii2; ++iii) {
  for (jjj=jj1; jjj<=jj2; ++jjj) {
    pixno = iii + (jjj * nc);
    sum=sum+bckimg_mb[pixno];
    num=num+1.;
  }}
  pixno = ii + (jj * nc);
  if (num > 0.) { scrimg[pixno] = sum / num; } else { scrimg[pixno]=0.; }
}}
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno = ii + (jj * nc);
  bckimg_mb[pixno] = scrimg[pixno];
}}
   #@# */


/* . . . . WITHOUT B image SUBTRACTION . . . . . . */
/* Create image of background fit and spectra of background fit. */
for (tso=3; tso<=7; ++tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (ii=0; ii<ecol; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    upper = tsx_find_image_row( 2, ii, tso, polyoff, TSP );
    nn8=0;
    for (jj=lower; jj<=upper; ++jj) {
      mm=jj-lower;
      if (tbck_mask[tso][mm] == 1) {
        pixno=ii+(jj*nc);
        xx8[nn8] = (double)mm;
        yy8[nn8] = image_nob[pixno];
        zz8[nn8] = 1.0;
        if ((maskimg[pixno] > 0.001)||(maskimg[pixno] < -0.001)) { zz8[nn8]=0.; }
        vv8[nn8] = image_nob_V[pixno];
        ++nn8;
      }
    }
    if (nn8 < 5) { printf("***error: spec.of.bck: nn8 < 5, nn8=%d .\n",nn8); exit(1); }
/* Sky fit (no B subtraction) */
    polyorder = 3;
    ok = tsx_skyfit( nn8, xx8, yy8, zz8, vv8, polyorder, &bck_xoff, bck_coef, 100., tso, ii );
    if (ok == 0) { printf("***error: fitting background polynomial.\n"); exit(1); }
    xv = ((exsp1+exsp2) / 2.) - bck_xoff;
    tspbck_nob[tso][ii] = cpolyval( polyorder+1, bck_coef, xv );
    for (jj=lower; jj<=upper; ++jj) {
      mm=jj-lower;
      xv = (double)mm - bck_xoff;
      yv = cpolyval( polyorder+1, bck_coef, xv );
      pixno=ii+(jj*nc);
      bckimg_nob[pixno] = yv; 
    }
  }
}



/* Mash orders of background image. */
for (tso=7; tso>=3; --tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (mm=0; mm<300; ++mm) { bckmash[mm]=0.; bckmcol[mm]=0.; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
  for (ii=20; ii<ecol-20; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      mm = jj-lower;
      bckmash[mm] = bckmash[mm] + bckimg_mb[pixno];
      bckmcol[mm] = bckmcol[mm] + 1.0;
    }
  }
/* Average and store. */
  for (mm=0; mm<mnum; ++mm) { bckpro_mb[tso][mm] = bckmash[mm] / bckmcol[mm]; }
}


/* Mash orders of background image. (using NO B SUBTRACTION image). */
for (tso=7; tso>=3; --tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (mm=0; mm<300; ++mm) { bckmash[mm]=0.; bckmcol[mm]=0.; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
/* Accumulate most columns. */
  for (ii=20; ii<ecol-20; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      mm = jj-lower;
      bckmash[mm] = bckmash[mm] + bckimg_nob[pixno];
      bckmcol[mm] = bckmcol[mm] + 1.0;
    }
  }
/* Average and store. */
  for (mm=0; mm<mnum; ++mm) { bckpro_nob[tso][mm] = bckmash[mm] / bckmcol[mm]; }
}




/* Mash orders using 'no B' subtraction image... */
for (tso=7; tso>=3; --tso) {
  if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  for (mm=0; mm<300; ++mm) { mash[mm]=0.; mcol[mm]=0.; }
  mnum = tsx_minwidth(ecol,tso,polyoff,TSP);    /* First find minimum width. */
  for (ii=20; ii<ecol-20; ++ii) {
    lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
    for (jj=lower; jj<(lower+mnum); ++jj) {
      pixno = ii+(jj*nc);
      wgt=1.;
      mm = jj-lower;
      mash[mm] = mash[mm] + (wgt * image_nob[pixno]);
      mcol[mm] = mcol[mm] + wgt;
    }
  }
  for (mm=0; mm<mnum; ++mm) { tpro_nob_wgt[tso][mm] = mash[mm] / mcol[mm]; }
}


/* Median filter and analyze profiles. */
for (tso=3; tso<=7; ++tso) {
  for (mm=0; mm<npro; ++mm) {
    fmbg_nob= tpro_nob[tso][mm] - bckpro_nob[tso][mm];
    arr8[mm] = tpro_nob[tso][mm];
    arr8b[mm]= fmbg_nob;
  }
  tsx_median_filter( 7, npro, arr8, mfarr8 );
  for (mm=0; mm<npro; ++mm) { tpro_nob_mf[tso][mm] = mfarr8[mm]; }
  tsx_median_filter( 7, npro, arr8b, mfarr8 );
  for (mm=0; mm<npro; ++mm) { tpro_fmbg_nob_mf[tso][mm] = mfarr8[mm]; }
}

/* Stats. */
for (tso=3; tso<=7; ++tso) {
  high_fmbg_nob_mf[tso]=-9999.;
  high_nob_mf[tso]    =-9999.;
  for (mm=8; mm<npro-8; ++mm) { 
    if (tpro_nob_mf[tso][mm]      > high_nob_mf[tso]     ) high_nob_mf[tso]     = tpro_nob_mf[tso][mm];
    if (tpro_fmbg_nob_mf[tso][mm] > high_fmbg_nob_mf[tso]) high_fmbg_nob_mf[tso]= tpro_fmbg_nob_mf[tso][mm];
  }
  if (high_nob_mf[tso]      > 99999) high_nob_mf[tso]      = 99999;
  if (high_fmbg_nob_mf[tso] > 99999) high_fmbg_nob_mf[tso] = 99999;
}



/* Create a object extraction window based on object mask. */
if ((autox)&&(bflag)) {
  tsx_autox( TOFF, npro, tpro_mb, bckpro_mb, tbck_mask, tobj_mask, &asp1, &asp2 );
  xv    = asp1 - TOFFI[3].ps_xoff;
  exsp1 = (( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
  xv    = asp2 - TOFFI[3].ps_xoff;
  exsp2 = (( cpolyval( TOFFI[3].polyorder+1, TOFFI[3].ps_coef, xv ) ));
  printf("tsx:auto: asp=%6.2f %6.2f\n",asp1,asp2);
  printf("tsx:auto:  sp=%6.2f %6.2f\n",exsp1,exsp2); 
  WriteSpectra=1;
}
if ((autox)&&(bflag == 0)) { printf("===WARNING: Auto extraction only works when B image is subtracted.\n"); }


/* Write out profiles. */
for (tso=3; tso<=7; ++tso) {
  sprintf(wrd,"%s%s-pro%d.tbl",tsxout,root,tso);
  outfu = fopen_write(wrd);
  fprintf(outfu,"| row | arcsec | flux       | bfit       |mask|rawpix|obj |fmb         | apix | flux_mb    | bfit_mb    | fmb_mb     |xw|\n");
/*                12345 12345678 123456789012 123456789012 1234 123456 1234 123456789012 123456 123456789012 123456789012 123456789012 12  */
  nopr_sum=0.;
  for (mm=0; mm<npro; ++mm) {
    tsx_relative_arcsec(tso,(double)mm,TOFF,&apix,&rarcs);
    fmbg_nob= tpro_nob[tso][mm] - bckpro_nob[tso][mm];
    fmbg_mb = tpro_mb[tso][mm]  - bckpro_mb[tso][mm];
    kk=0;
    if (asp1 > -800.) {
      if ((rarcs >= asp1)&&(rarcs <= asp2)) { kk=1; }
    }
    fprintf(outfu," %5d %8.3f %12.5e %12.5e %4d %6d %4d %12.5e %6.2f %12.5e %12.5e %12.5e %2d \n",
          mm,rarcs,tpro_nob[tso][mm],bckpro_nob[tso][mm],tbck_mask[tso][mm],0,tobj_mask[tso][mm],fmbg_nob,apix,
          tpro_mb[tso][mm],bckpro_mb[tso][mm],fmbg_mb, kk );
    if (kk) { 
      nopr[tso][mm] = tpro_mb[tso][mm]; 
      nopr_sum = nopr_sum + nopr[tso][mm]; 
    } else { nopr[tso][mm]=0.; }
  }
  fclose(outfu);

/* Normalized profiles. */
  if (nopr_sum < 0.0001) { 
    for (mm=0; mm<npro; ++mm) { nopr[tso][mm] = 0.; }
  } else {
    for (mm=0; mm<npro; ++mm) { nopr[tso][mm] = nopr[tso][mm] / nopr_sum; }
  }

}

/* #@# copy of image_mb ..
for (ii=0; ii<nc; ++ii) {
for (jj=0; jj<nr; ++jj) {
  pixno=ii+(jj*nc);
  scrimg[pixno] = image_mb[pixno];
}}
tsx_wout("image_mb.fits",scrimg,nc,nr);
*/


/* Find an appropriate hot star (unless given by user). */
if (strcmp(HOTroot,"")==0) {
  tsx_find_hotstar( tsxout, root, TLS, HOTroot );
}



/* Set for Re-Bin image. */
RBrow=0; 


/* Create spectra of object. */
if ((WriteSpectra)&&(exsp1 > 0.)&&(exsp2 > exsp1)) {

  printf("Extracting object spectrum, asp1=%0.2f  asp2=%0.2f  (sp1=%0.2f  sp2=%0.2f for TSpec Order=3).\n",asp1,asp2,exsp1,exsp2);
  FoundAC=0;
  for (tso=3; tso<=7; ++tso) {

    if (tso > 3) {
      xv = exsp1 - TOFF[tso].ps_xoff;
      yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
      isp1 = cnint(( exsp1 + yv ));
      rsp1 =      (( exsp1 + yv ));
      xv = exsp2 - TOFF[tso].ps_xoff;
      yv = cpolyval( TOFF[tso].polyorder+1, TOFF[tso].ps_coef, xv );
      isp2 = cnint(( exsp2 + yv ));
      rsp2 =      (( exsp2 + yv ));
    } else {
      isp1=cnint(exsp1); isp2=cnint(exsp2);
      rsp1=exsp1; rsp2=exsp2;
    }
    if (tso == 7) { ecol=nc/2; } else { ecol=nc; }

/* Re-Binned Image.. */
    for (ii=0; ii<ecol; ++ii) {
      lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
      upper = tsx_find_image_row( 2, ii, tso, polyoff, TSP );
      yv0 = tsx_find_real_image_row( ii, tso, polyoff, TSP );
      for (offT=1.0; offT<=((double)(upper - lower)); offT=offT+1.) {
        sum=0.; wsum=0.; vsum=0.;
        for (offy=offT; offy<(offT+1.0); offy=offy+0.01) {
          yv = yv0 + offy; 
          jj = cnint(yv);
          pixno = ii + (jj*nc);
/* inc. bckimg subt. -tab 16sep2014 */
          sum = sum + (image_mb[pixno] - bckimg_mb[pixno]);   
          vsum= vsum+ image_mb_V[pixno];
          wsum= wsum+ 1.0;
        }
        RBrow = cnint(offT) + ((4 - (tso-3)) * 200);
        pp = ii + (RBrow * nc);
        RBimage[pp] = sum / wsum;
        RBimageV[pp]= vsum/ wsum;
      }
    }


    for (ii=0; ii<ecol; ++ii) {
      lower = tsx_find_image_row( 1, ii, tso, polyoff, TSP );
      upper = tsx_find_image_row( 2, ii, tso, polyoff, TSP );


/* Find CRs in object window. */
/* work here ...  */
/* work here ... */
/* work here ... */
      if (tso == 999) { 
      for (jj=lower; jj<=upper; ++jj) { mm=jj-lower;
        if ((mm >= isp1)&&(mm <= isp2)) {
          pixno=ii+(jj*nc);
          ii1=ii-1; ii2=ii+1; 
          if (ii1 < 0   ) ii1=0; 
          if (ii2 > nc-1) ii2=nc-1;
          jj1=jj-2; jj2=jj+2; 
          if (jj1 < lower) jj1=lower; 
          if (jj2 > upper) jj2=upper;
          bbb=0;
          for (iii=ii1; iii<=ii2; ++iii) {
          for (jjj=jj1; jjj<=ii2; ++jjj) {
            ppp=iii+(jjj*nc);
            if (ppp != pixno) {
              mmm=jjj-lower;
              if (nopr[tso][mmm] > 0.) {
/*
                arr4[bbb] = (image_mb[ppp] - bckimg_mb[ppp]) / nopr[tso][mmm];
*/
                arr4[bbb] = (image_mb[ppp] - bckimg_mb[ppp]);
                ++bbb;
              }
            }
          }}
          mean = cfind_mean_rms( bbb, arr4, &rms);
          if ((rms > 0.)&&(nopr[tso][mm] > 0.)) {
/*
            dev = ((image_mb[pixno] - bckimg_mb[pixno]) / nopr[tso][mm]) - mean;
*/
            dev = (image_mb[pixno] - bckimg_mb[pixno]) - mean;
            if (dev/rms > 3.) {
              printf("#@# CR in object: %d  ii=%d  dev=%f  rms=%f  dev/rms=%f\n",tso,ii,dev,rms,dev/rms);
              image_mb[pixno] = mean + bckimg_mb[pixno];
              cpauseit();
            }
          }
        }
      }
      }

/* Extract spectral point using 'Re-Binned Image' technique..  -tab 16sep2014 .. */
      yv0 = tsx_find_real_image_row( ii, tso, polyoff, TSP );
      sum=0.; wsum=0.; vsum=0.; rawrow=0.;
      for (offy=rsp1; offy<=rsp2; offy=offy+0.01) {
        yv = yv0 + offy;
        jj = cnint(yv);
        pixno = ii + (jj*nc);
        sum = sum + ( image_mb[pixno] - bckimg_mb[pixno] );
        vsum= vsum+ ( image_mb_V[pixno] );
        rawrow = rawrow + (double)jj;
        wsum= wsum+ 1.0;
      }
      sum = sum * 0.01;     /* since we are sampling every hundredth pixel */
      vsum= vsum* 0.01;
      num = wsum;

      if (num < 1.) { printf("***error: spec.of.obj: num < 1.\n"); exit(1); }
      tspobj[tso][ii] = sum;
      if (num  > 0.) { tsprow[tso][ii] = cnint((rawrow / num)); }
      if (vsum > 0.) { tsperr[tso][ii] = sqrt(vsum); tsps2n[tso][ii] = sum / sqrt(vsum); }
                else { tsperr[tso][ii] = 0.;         tsps2n[tso][ii] = 0.;               } 
    }
  }

/* NOTE: Use polyorder = 3 for all orders for the time being... -tab 04dec2013... */
  for (oo=0; oo<9; ++oo) { tsns_coef[oo]=0.; }

/* Write out spectra for each order. */
  for (tso=3; tso<=7; ++tso) {

    if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
  
/* Compute wavelength scale from sky line ID file. */
    polyorder = 3;  ns_shift=0.;
    tsx_load_tsns( &tsns_n, tsns_x, tsns_y, &ns_shift, verbose, tso, tsxdir, ecol, tspbck_nob );

/* Load and fit. */
    for (kk=0; kk<tsns_n; ++kk) {
      xx8[kk] = tsns_x[kk] + ns_shift;
      yy8[kk] = tsns_y[kk];
      zz8[kk] = 1.0;
    }
    tsns_xoff = 0.;
    ok = GJ_polyfit( tsns_n, xx8, yy8, zz8, polyorder, 1, &tsns_xoff, tsns_coef );
    if (ok == 0) { printf("***error: fitting sky line ID wavelength scale polynomial.\n"); exit(1); }
    GJ_polyfit_residuals( tsns_n, xx8, yy8, zz8, polyorder, tsns_xoff, tsns_coef, &high, &wrms, &wppd );

/* Dispersion. */
    hidisp=0.; lodisp=1.e+30;
    for (ii=ecol-1; ii>=0; ii=ii-20) {
      icol = (ecol-1) - ii;         /* inverse column */
      xv = (double)(icol) - tsns_xoff;
      disp =  cpolyval(polyorder+1,tsns_coef,xv+1.) - cpolyval(polyorder+1,tsns_coef,xv);
      if (disp > hidisp) hidisp=disp;
      if (disp < lodisp) lodisp=disp;
    }
    disp = (hidisp + lodisp) / 2.;
    printf("NOTE: sky wavelength fit: tso=%d N=%3d  disp=%0.2f(%0.2f-%0.2f)A/px  high=%5.2f(px) wrms=%0.3f(px) [sky shift=%5.2f px]\n",
        tso,tsns_n,disp,lodisp,hidisp,high/disp,wrms/disp,ns_shift);

/* 'Re-Binned' image wavelengths. */
    for (ii=0; ii<ecol; ++ii) {
      icol = (ecol-1) - ii;         /* inverse column */
      xv = (double)(icol) - tsns_xoff;
      wave =  cpolyval( polyorder+1, tsns_coef, xv );
      jj = 4 - (tso - 3);
      pixno = ii + (jj * nc);
      RBwave[pixno] = wave / 10000.;
    }

/* Set ecol. */
    if (tso == 7) { ecol=nc/2; } else { ecol=nc; }

/* Read HSy[] if HOTroot exists. */
    HSn=0; for (ii=ecol-1; ii>=0; --ii) { HSy[HSn]=1.; ++HSn; }
    if (strcmp(HOTroot,"") != 0) {
      HSn=0; pp=0; 
      sprintf(wrd,"%s/%s-sp%d.tbl",tsxout,HOTroot,tso);
      if (FileExist(wrd)) {
        infu = fopen_read(wrd);
        while (fgetline(line,infu)) {
          if (line[0]=='|') { 
            pp = cindex(line,"VsumD"); 
          } else {
            if (pp > 0) { 
              substrcpy_terminate(line,pp-1,clc(line),wrd2,0); 
              HSy[HSn]=GLV(wrd2,1); 
              ++HSn; 
            }
          }
        }
        fclose(infu);
        ok=0;
        for (ii=0; ii<HSn; ++ii) { if (HSy[ii] > 0.) ok=1; }
        if (ok == 0) {
          printf("===warning: found '%s', but it is not the correct temperature for atmospheric correction.\n",wrd);
        } else { FoundAC=1; }
      } else {
        printf("===warning: cannot find '%s'.\n",wrd);
      }
    }

/* Load Vx,y.  Here we load Vx,Vy which will be used to create a 'divisble-spectrum' if this is
   a Vega-like star.  Here we also create 'Vac' which uses a previously created divisible-spectrum
   from another exposure of a Vega-like star.  If this star turns out to not have the correct
   temperature, then the 'VsumD' column (VyD) will be -1. */
    Vn=0;
    for (ii=ecol-1; ii>=0; --ii) {
      icol = (ecol-1) - ii;         /* inverse column */
      xv = (double)(icol) - tsns_xoff;
      if (polyorder > 0) { wave =  cpolyval( polyorder+1, tsns_coef, xv ); } else { wave = 0.; }
      Vx[Vn] = wave / 10000.;
      Vy[Vn] = tspobj[tso][ii];
      VyD[Vn]= -1.;
      if (HSy[Vn] > 0.) { rr=HSy[Vn]; } else { rr=1.; }
      if (FoundAC == 0) { Vac[Vn]=0.; } else {
        Vac[Vn]= tspobj[tso][ii] / rr;
      }
      ++Vn;
    }


/* Create divisible spectrum if this is a Vega-like star. */
    if ((EffTemp > 8900.)&&(EffTemp < 10300.)&&(vega)) {
      printf("This is a Vega-like star, generate divisible spectrum...\n");
      tsx_vega( tso, Vn, Vx, Vy, VCn, VCx, VCy, Van, Va1, Va2, VyD );
    } else {
      printf("This is NOT a Vega-like star, generate divisible spectrum anyway...\n");
      tsx_vega( tso, Vn, Vx, Vy, VCn, VCx, VCy, Van, Va1, Va2, VyD );
    }

/* Write spectrum. */
    if (tso == 7) { ecol=nc/2; } else { ecol=nc; }
    sprintf(wrd,"%s%s-sp%d.tbl",tsxout,root,tso);
    outfu = fopen_write(wrd);
    fprintf(outfu,"| pix | wave    | sum        | s2n        | backgnd    | cps        | flux       | col  |backgnd_nob | error    |row | sumac      | VsumD      |\n");
/*                  12345 123456789 123456789012 123456789012 123456789012 123456789012 123456789012 123456 123456789012 1234567890 1234 123456789012 123456789012 */
    Vn=0;
    for (ii=ecol-1; ii>=0; --ii) {
      icol = (ecol-1) - ii;         /* inverse column */
      xv = (double)(icol) - tsns_xoff;
      if (polyorder > 0) { wave =  cpolyval( polyorder+1, tsns_coef, xv ); } else { wave = 0.; }
      fprintf(outfu," %5d %9.5f %12.5e %12.5e %12.5e %12.5e %12.5e %6d %12.5e %10.3e %4d %12.5e %12.5e \n",
            icol,wave/10000.,tspobj[tso][ii],tsps2n[tso][ii],tspbck_mb[tso][ii],tspobj[tso][ii]/exptime,
            tspobj[tso][ii]/exptime,ii,tspbck_nob[tso][ii],tsperr[tso][ii],tsprow[tso][ii],Vac[Vn],VyD[Vn]);
      ++Vn;
    }
    fclose(outfu);

  
  }    /*  for (tso=3; tso<=7; ++tso)  ... */

}      /* if ((exsp1 > 0)&&(exsp2 > 0)) ... */


/* Save ReBinned processed image, if requested. */
if (rbi) {
  sprintf(wrd,"%s%s-rbi.fits",tsxout,root); tsx_wout(wrd,RBimage ,nc,nr);
  sprintf(wrd,"%s%s-rbv.fits",tsxout,root); tsx_wout(wrd,RBimageV,nc,nr);
  sprintf(wrd,"%s%s-rbw.fits",tsxout,root); tsx_wout(wrd,RBwave  ,nc, 5);
}

/* Write to list file (or stdout). */
misc_pad_blanks(utshut  ,19,wrd1);
misc_pad_blanks(object  ,14,wrd2);
misc_pad_blanks(root    ,20,wrd3);
if (exptime > 0.) { rr1 = (high_nob_mf[3] - high_fmbg_nob_mf[3]) / exptime; } else { rr1 = 0.; }
bcps = cnint(rr1);
if (bcps > 9999) bcps=9999;
if (bcps < -999) bcps=-999;
/* Type? */
strcpy(xtype,"unkn");
if (bcps > 50)                                  { strcpy(xtype,"flat"); } else {
if ((high_fmbg_nob_mf[3] < 5.)&&(averge < 20.)) { strcpy(xtype,"dark"); } else {
if (high_fmbg_nob_mf[3] > 10.)                  { strcpy(xtype,"objt"); }      }}

/* Header. */
if (listfu == stdout) {
fprintf(listfu,"|  UTSHUT           | OBJECT       | ra      | dec     | air | nc | nr |averge|d|t| expt  |hi3wb|hi3mb|bcps|type| root               | los | EffTemp | file_directory \n");
}
fprintf(listfu," %s %s %9.5f %9.5f %5.3f %4d %4d %6.1f %d %d %7.2f %5d %5d %4d %s %s %5.1f %9.2f %s\n",
   wrd1,wrd2,ra_deg,dec_deg,airmass,nc,nr,averge,detflag,typflag,exptime,
   cnint(high_nob_mf[3]),cnint(high_fmbg_nob_mf[3]),bcps,xtype,wrd3,los*3600.,EffTemp,fdir);

/* Write PS plots. */
if (noplot == 0) {
  sprintf(wrd,"tsxplot %s x=arcsec y=flux_mb -pro -ps object='%s'",root,object); misc_system(wrd);
  if ((WriteSpectra)&&(exsp1 > 0.)&&(exsp2 > exsp1)) {
    sprintf(wrd,"tsxplot %s x=wave y=sum z=error -sp -ps object='%s'",root,object); misc_system(wrd);
    if (strcmp(HOTroot,"") != 0) {
      sprintf(wrd,"tsxplot %s x=wave y=sumac z=error -sp ps=%sac object='%s'",root,root,object); misc_system(wrd);
    }
  }
}


/* #@# 
strcpy(wrd,"test.fits");
printf("writing '%s'.\n",wrd);
tsx_wout(wrd,image,nc,nr);
strcpy(wrd,"test2.fits");
printf("writing '%s'.\n",wrd);
tsx_wout(wrd,image2,nc,nr);
   #@# */

/* Free. */
free(image_nob);   image_nob  = NULL;
free(image_nob_V); image_nob_V= NULL;
free(image_mb);    image_mb   = NULL;
free(image_mb_V);  image_mb_V = NULL;
free(bimage);      bimage     = NULL;
free(bimage_V);    bimage_V   = NULL;
free(bckimg_nob);  bckimg_nob = NULL;
free(bckimg_mb);   bckimg_mb  = NULL;
free(RBimage);     RBimage    = NULL;
free(RBimageV);    RBimageV   = NULL;
free(RBwave);      RBwave     = NULL;


}
/* - - - - - - END OF BIG LOOP - - - - - - - */

/* Free. */
free(maskimg);     maskimg    = NULL;
free(darkimg);     darkimg    = NULL;
free(flatimg);     flatimg    = NULL;
free(TLS); free(TYC);

for (ii=0; ii<maxff; ++ii) { free(ff_a[ii]); free(ff_b[ii]); }
if (listfu != stdout) fclose(listfu);
return(0);
}
